#!/usr/bin/perl

################################################################################
# This program gets as input a Rfam like directory
# , containing subdirectories of multiple alignments,
# Then maps the consensus structure of each alignment ot the its sequences
# For the consensus structure two option is available:
#	a) using the Rfam like .cons file in each directory
#	b) run PETfold to produce consensus structure
# Then a constraint folding is done using RNAfold, where some sequences might filtered out beacuse of not folding with compatiable to the consensus
# Output:
#		For each alignment, a fasta file containg the filtered sequences and their corresponding secondary structure with tag #S is produced
#		this output can be used as input by alignmnet2gspan.pl
################################################################################

#use feature ':5.10';
use strict 'vars';
use warnings;
use Getopt::Long;
use Pod::Usage;
use List::Util qw/ min max /;

#use POSIX qw/ceil floor/;
use Cwd qw/abs_path getcwd/;

use FindBin;                        # locate this script
use lib "$FindBin::Bin/../lib/";    # use the parent directory

use Consensus;

#use lib ("../lib/");
use StructureLibrary::Tools;
use StructureLibrary::Sequence;
use MultipleAlignment::Tools;
use MultipleAlignment::PETfold;
use lib "$FindBin::Bin/../";
##load config file
use Config::AlignClustConfig;

###############################################################################
# GLOBAL VARIABLES
###############################################################################
my $CURRDIR             = getcwd;
my $RESULTDIR           = "/Folded_Seqs/";                      #updated below
my $RUN_PETFOLD         = "$FindBin::Bin/run_PETfold.pl";
my $RUN_ALIGNMENT_GSPAN = "$FindBin::Bin/alignment_gspan.pl";
my $STOCKHOLM_SUFFIX    = ".stk";                               #".stockholm";
my $CONSENSUS_SUFFIX    = ".cons";
my $CLUSTAL_SUFFIX      = ".clustal";
my $ALN_FASTA_SUFFIX    = ".aln.fasta";
my $REP_FASTA_SUFFIX    = ".rep_fa";
my %RFAMSEED            = ();

#my %RFAMFILTERED = ();

=head1 NAME


alignment_to_secondary_structures.pl -input-dir "alignments directory" 

=head1 SYNOPSIS

alignment_to_secondary_structures.pl

Options:

		
        -help    		brief help message
        -man     		full documentation
        -input-aln-file		alignment file <mandatory>
        -input-cons-file		corresponding cons file for clustal alignments 
        -stdout  		redirect result fasta to stdout
        -local-config		if a local configuration is to be used, path to the local configuration file
        -debug	 		debug mode
        -result-dir	 	output directory for results
	-tmp-dir		path of temporaty directory
	-keep-tmp		if this flag is set, the temporary directory is not deleted after execution
        -cutoff-ratio-consensus=i
        -cutoff-ratio-MFE=i  
        -use-petfold
        -mfe 			Use MFE of each sequence as encoded graph,\
                     	this option ignores alignment and evolution information\
                      	and is mainly for evaluation purposes
        -only-representative    Only encode the given representative as graph and ignore other seqeunces 
                      
        -read-stockholm
		-read-aln-fasta		file ending on .aln.fasta storing alignments in FASTA format
        -threshold-reliablity=f

=head1 DESCRIPTION

=cut

###############################################################################
# PARSE COMMAND LINE OPTIONS
###############################################################################
print STDOUT ( join( "|", @ARGV ), "\n" );

# command line options
my (
	$i_help,           $i_man,          $i_aln_file,   $i_cons_file, $i_stdout,
	$i_debug,          $i_local_config, $i_result_dir, $i_tmp_dir,   $i_cutoff_cons,
	$i_cutoff_mfe,     $i_petfold,      $i_stockholm,  $i_clustal,   $i_aln_fasta,
	$i_threshold_reli, $i_job_id, $i_mfe, $i_path_graph, $i_path_abstract_graph, $i_max_selection_size
);
$i_cutoff_cons    = 0;
$i_cutoff_mfe     = 0;
$i_threshold_reli = 0.001;
my $options = GetOptions(
	"help"                     => \$i_help,
	"man"                      => \$i_man,
	"stdout"                   => \$i_stdout,
	"debug"                    => \$i_debug,
	"local-config=s"           => \$i_local_config,
	"input-aln-file=s"         => \$i_aln_file,
	"input-cons-file=s"        => \$i_cons_file,
	"result-dir=s"             => \$i_result_dir,
	"tmp-dir=s"                => \$i_tmp_dir,
	"cutoff-ratio-consensus=i" => \$i_cutoff_cons,
	"cutoff-ratio-MFE=i"       => \$i_cutoff_mfe,
	"use-petfold"              => \$i_petfold,
	"mfe"                      => \$i_mfe,
	"path-graph"               => \$i_path_graph,
	"path-abstract-graph"      => \$i_path_abstract_graph,
	"max-selection-size=i"     => \$i_max_selection_size,
	"read-stockholm"           => \$i_stockholm,
	"read-clustal"             => \$i_clustal,
	"read-aln-fasta"           => \$i_aln_fasta,
	"threshold-reliablity=f"   => \$i_threshold_reli,
	"job-id=s"                 => \$i_job_id
);

pod2usage( -exitstatus => 1, -verbose => 1 ) if $i_help;
pod2usage( -exitstatus => 0, -verbose => 2 ) if $i_man;
($options) or pod2usage(2);

if ($i_local_config) {
	( -e $i_local_config )
	  or pod2usage("error: local configuration file does not exist: $i_local_config\n");
}

#foreach keys print "alignment_to_secondary_structures.pl " . $options ."\n" if ($i_debug);
($i_aln_file) or pod2usage("error: -input-file alignmnet is not given.\n");

( $i_petfold || $i_stockholm || $i_aln_fasta || $i_cons_file )
  or pod2usage(
"error: consensus struct is not defined. It can be computed by petfold or retreived from stockholm or cons struct file.\n"
  );

if ($i_aln_file) {
	( -e $i_aln_file ) or pod2usage("error: no such input alignmnet file: $i_aln_file\n");
}

if ($i_cons_file) {
	( -e $i_cons_file ) or pod2usage("error: no such input alignmnet file: $i_cons_file\n");
}

if ($i_result_dir) {
	( -e $i_result_dir ) or pod2usage("error: no such directory: $i_result_dir\n");
	$RESULTDIR = abs_path($i_result_dir) . $RESULTDIR;
}
else {
	$RESULTDIR = $CURRDIR . $RESULTDIR;
}

( $i_cutoff_cons >= 0 and $i_cutoff_cons <= 100 )
  or
  pod2usage("error: cutoff ratio consesnusis percentage and must be an integer between 0 and 100");
( $i_cutoff_mfe >= 0 and $i_cutoff_mfe <= 100 )
  or pod2usage("error: cutoff ratio mfe is percentage and must be an integer between 0 and 100");

if ( $i_debug && $i_stdout ) {
	pod2usage("error: at most one of debug and stdout switches can be on");
}

my $activated_formats = 0;
if ($i_clustal)   { $activated_formats++; }
if ($i_stockholm) { $activated_formats++; }
if ($i_aln_fasta) { $activated_formats++; }
( $activated_formats == 1 )
  or pod2usage("error: exactly one of reading format should be turned on");

if ($i_clustal) {
	print STDERR
"Warning:\tread_clustal option is experimental, \n\t\tclustal files would be treated as stockholm files without a consensus strcuture\n";
	($i_petfold)
	  or pod2usage(
		"Error read_clustal only works with pet fold prediction, use_petfold would be mandatory\n");
}

( defined $i_job_id and $i_job_id ) or $i_job_id = "local";

my $tmpdir = $i_tmp_dir;
#create parent tmp dir if needed
unless ( -d $tmpdir ) {
	mkdir $tmpdir, 0777
	  or die "ERROR: cannot create subdirectory $tmpdir in temporary directory!:\n\t$!\n";
}
## create unique tmp directory
$tmpdir .= "/tmp.$i_job_id/";
unless ( -d $tmpdir ) {
	mkdir $tmpdir, 0777
	  or die "ERROR: cannot create subdirectory $tmpdir in temporary directory!:\n\t$!\n";
}

###############################################################################
# Overwrite global CONFIG with local configuration
###############################################################################
my $CONFIG_REF;
if ($i_local_config) {
	$CONFIG_REF = Config::AlignClustConfig::load_config($i_local_config);
}
else {
	$CONFIG_REF = Config::AlignClustConfig::load_config();
}

my %CONFIG = %{$CONFIG_REF};
Consensus::LOAD_CONFIG($i_local_config);

###############################################################################
# end handler for temporary directory
# adds an error handler that deletes the directory in case of error
# SIGUSR{1/2} are sent by the sge prior to the uncatchable SIGKILL if the
# option -notify was set
###############################################################################
$SIG{'INT'}  = 'end_handler';
$SIG{'TERM'} = 'end_handler';
$SIG{'ABRT'} = 'end_handler';
$SIG{'USR1'} = 'end_handler';
$SIG{'USR2'} = 'end_handler';

sub end_handler {
	print STDERR "signal ", $_[0], " caught, cleaning up temporary files\n";

	# change into home directory. deletion of the temporary directory will
	# fail if it is the current working directory
	chdir();
	File::Temp::cleanup();
	die();
}

###############################################################################
# EXECUTION CODE
###############################################################################
#
#my $string = "((((...((.))...(())...((.))...((..))(..)..))))";
#print $string."\n";
### remove lonely base-pairs
#$string =~ s/\.\(\./\.\.\./g;
#$string =~ s/\)\(\./\)\.\./g;
#$string =~ s/\.\)\./\.\.\./g;
#$string =~ s/\.\)\(/\.\.\(/g;
## remove base-pairs that result in a 0,1,2-sized hairpin loop
#$string =~ s/\(\)/../g;
#$string =~ s/\(\.\)/\.\.\./g;
#$string =~ s/\(\.\.\)/\.\.\.\./g;
#print $string."\n";
#die "testing\n";

#------

my $cons_struct_file;
my $stk_file;
## get consensus file, unless using petfold
my $current_cons_file = "";

if ($i_stockholm)    #read stockholm input
{
	$stk_file = $i_aln_file;
}
elsif ($i_clustal)    #clustal format is transformed first to stk format
{
	my $clustal_file = $i_aln_file;
	my @parts        = get_path_base_suffix($clustal_file);

	$stk_file = $parts[0] . "/" . $parts[1] . $STOCKHOLM_SUFFIX;
	print "$stk_file\n" if ($i_debug);
	MultipleAlignment::Tools::convert_clustal_to_stk( $clustal_file, $stk_file );
}

if ( !$i_petfold && ( $i_stockholm || $i_clustal ) ) {
	$current_cons_file = stockholm_to_cons( $stk_file, abs_path($stk_file) . $CONSENSUS_SUFFIX );
}
elsif ( !$i_petfold && $i_cons_file ) {
	$current_cons_file = $i_cons_file;
}

############################################################
## get information about the path, file_base name, etc.
## if i_petfold: stk file, others: consensus file
my @parts_cons;
if ( $i_stockholm || $i_clustal ) {
	@parts_cons = get_path_base_suffix($stk_file);
}
else    #aln-fasta-file is directly used and no stk-file is generated
{
	@parts_cons = get_path_base_suffix($i_aln_file);
}

my $file_path      = $parts_cons[0];
my $file_name      = $parts_cons[1];
my $file_extension = $parts_cons[2];

##filter rfam id
my $rfam_id = "";
if ( $file_name =~ /(RF\S+)\.alignment/ ) {
	$rfam_id = $1;
}
else {
	$rfam_id = $file_name;
}

print STDOUT "getting structures for $file_name\n" if ($i_debug);

## fasta alignmnet file is needed
my $aln_fasta_file = "$file_path/$file_name" . $ALN_FASTA_SUFFIX;
if ( $i_stockholm || $i_clustal )    #if input is stockholm, first creat the fasta alignment
{
	stockholm_to_aln_fasta( $stk_file, $aln_fasta_file );
}
elsif ($i_aln_fasta) {
	$aln_fasta_file = $i_aln_file;
}
die "Error fasta alignmnet $aln_fasta_file not found\n" if ( !-e $aln_fasta_file );

###########################################################

## Map consensus strcuture to sequences
{

	my @structure_contents = ();
	my $headerline =
	    "NAME\tRATIO_REMAINING_CONSTRAINT\tSTRUCTURE_CONSTRAINT\t"
	  . "STRUCTURE_CONSTRAINT_MFE\tRATIO_CONSTRAINT_MFE\t"
	  . "ALL_ENSEMBLE_ENERGY\tCONSTRAINT_ENSEMBLE_ENERGY\t"
	  . "CONSTRAINT_MFE_ENERGY\tCONSTRAINT_PROBABILITY\t"
	  . "CONSTRAINT_MFE_PROBABILITY\t"
	  . "REL_DIFF_CONSTRAINT_P_AND_CONSTRAINT_MFE_P";
	push( @structure_contents, $headerline );

	my $consensus;
	if ($i_petfold) {
		$consensus = return_consensus_struct($aln_fasta_file);    #calculate cons structure
	}
	else {

		#extract cons structure
		$consensus = return_consensus_struct( $aln_fasta_file, $current_cons_file );
	}

	my ( $tags_aref, $inorder_seqs_aref );
	my ( $seqs_href, $markup_href );

	## make an inorder list
	if ( $i_stockholm || $i_clustal ) {

# TODO reading stockholm file multiple times is inefficient! It could be merged with i_aln_fasta case
		( $seqs_href, $markup_href, $tags_aref ) =
		  MultipleAlignment::Tools::read_stockholm( $stk_file, 1 );    #only GC is enough
		my @inorder_seqs = map { ${$seqs_href}{$_} } @{$tags_aref};
		$inorder_seqs_aref = \@inorder_seqs;
	}
	else    #case $i_aln_fasta
	{       # TODO  double check functionality for this case  !

		my $rest_headers_href;
		( $seqs_href, $tags_aref, $rest_headers_href ) =
		  StructureLibrary::Sequence::read_fasta_file($aln_fasta_file);
		my @inorder_seqs = map { ${$seqs_href}{$_} } @{$tags_aref};
		$inorder_seqs_aref = \@inorder_seqs;
	}
	
	if ($i_mfe) {
		$consensus = "." x length($consensus)  
	}
	
	if ($i_debug) {

		# 		print STDOUT "tags are: " . join (", " , @{$tags_aref}) . "\n" ;
		# 		print STDOUT "inorder seqs are: " . join (", " , @{$inorder_seqs_aref}) . "\n";
		print STDOUT $consensus . " <-- cons struct\n";
	}
	
	next if ( @{$tags_aref} == 0 );

    ##=========
    ## Find out the representative sequence from the repfa file
	my $rep_fasta_file = "$file_path/$file_name" . $REP_FASTA_SUFFIX;
	print "rep_fasta_file: $rep_fasta_file\n";
	my ($rep_seqs_href, $rep_tags_aref, $rep_rest_headers_href);
	($rep_seqs_href, $rep_tags_aref, $rep_rest_headers_href ) =
	  StructureLibrary::Sequence::read_fasta_file($rep_fasta_file);
	print "rep_fasta_id: ", @{$rep_tags_aref}, "\n";
    
    ##=========
    ## Set the sequences to iterate
	
	my $select_count = 1;

	my $seq_tags_to_iterate = $rep_tags_aref;
	
       # select other sequences, one can shuffle first, but that would make the evaluation inconsistent
        my $num_seqs_needed = min(scalar($i_max_selection_size), scalar(@{$tags_aref}));
        print "num_seqs_needed: $num_seqs_needed, $i_max_selection_size\n";
        while (@{$seq_tags_to_iterate} < $num_seqs_needed) {
                 my $cur_tag = @{$tags_aref}[ $select_count - 1 ];
                 print "$select_count\n" ;
                 $select_count += 1;
                 if ($cur_tag eq @{$rep_tags_aref}[0]) {
                        print "skip rep $cur_tag\n";
                        print @{$rep_tags_aref}[0] . "\n";
                 }
                 else {
                        push(@$seq_tags_to_iterate, $cur_tag);
                 }
                 print join(" ", @$seq_tags_to_iterate);
                 print "\n";
        }

	
	
	
	## iterate through each element in alignment
  ELEMENT_LOOP: foreach my $elem_count ( 1 .. @{$seq_tags_to_iterate} ) {
		my $elem = @{$seq_tags_to_iterate}[ $elem_count - 1 ];
		if ($i_debug) {
			print STDOUT "\nElement: $elem\n";
			print STDOUT "@{$inorder_seqs_aref}[$elem_count-1]<--seq\n";
		}

		#always run once with cutoff zero
		my @cutoff_cons_arr = (0);
		my @cutoff_mfe_arr  = (0);
		if ( $i_cutoff_cons != 0 or $i_cutoff_mfe != 0 ) {
			push @cutoff_cons_arr, $i_cutoff_cons;
			push @cutoff_mfe_arr,  $i_cutoff_mfe;
		}
		for my $i_cut ( 0 .. $#cutoff_cons_arr ) {
			print STDOUT "cutoff cons:"
			  . $cutoff_cons_arr[$i_cut] . " mfe:"
			  . $cutoff_mfe_arr[$i_cut] . "\n"
			  if ($i_debug);

			my $elem_line = Consensus::project_constraint_and_fold_seq(
				$consensus,              @{$inorder_seqs_aref}[ $elem_count - 1 ],
				$elem,                   $elem_count,
				$rfam_id,                $cutoff_cons_arr[$i_cut],
				$cutoff_mfe_arr[$i_cut], $i_stdout,
				$RESULTDIR,              $tmpdir
			);
			if ( $elem_line ne "" ) {    # If cutoff thresholds are passed
				if ($i_debug) {
					my @elem_arr = split( /\t/, $elem_line );
					print STDOUT $elem_arr[5] . " <--folded struct\n";
				}
				push( @structure_contents, $elem_line );
				system("rm -R $tmpdir/*.fa");
				system("rm -R $tmpdir/*.fasta");
			}
		}
	}
	system("rm -R $tmpdir/*");

	#TODO following needed:
	# if ($i_job_id ne "local") {
	#	rmdir $tmpdir or die "ERROR: cannot remove subdirectory in temporary directory!:\n\t$!\n";
	# }
	StructureLibrary::Tools::write_file_from_array( abs_path( $parts_cons[0] . "/$rfam_id.struct" ),
		\@structure_contents );
	## .folded.fasta file should be generated by Consensus::project_constraint_and_fold_seq
	my $folded_fasta = "$RESULTDIR/$rfam_id.folded.fa";
	if ( -e $folded_fasta ) {
		my $alignment_gspan_tmp_dir = $tmpdir . "/alignment_gspan/";
		unless ( -d $alignment_gspan_tmp_dir ) {
			mkdir $alignment_gspan_tmp_dir, 0777
			  or die "ERROR: cannot create subdirectory $tmpdir in temporary directory!:\n\t$!\n";
		}
		my $debug_param = "";
		$debug_param = "-debug" if $i_debug;
		my $alignment_gspan_call =
		    "$RUN_ALIGNMENT_GSPAN -fasta $folded_fasta -group 100 $debug_param "
		  . "-o $RESULTDIR/../GSPAN/ -tmp $alignment_gspan_tmp_dir $CONFIG{OPTIONS_ALN_GSPAN} ";
		if ($i_path_graph) { # Only add path graph no structure
			$alignment_gspan_call .= "-nostr ";
		}
		if ($i_path_abstract_graph) { # Only add path graph no structure
			$alignment_gspan_call .= "-abstr ";
		}
		
		print STDOUT "Calling: $alignment_gspan_call \n";
		system($alignment_gspan_call);
	}
	else {
		die "ERROR: $folded_fasta file not found!";
	}
	if ($i_debug) {
		print STDOUT "Rfamid is $rfam_id\n";
		print STDOUT "Written structure information to: " . $parts_cons[0] . "/$rfam_id.struct\n";
	}

}

##############
#map_cons_fold_seq($consensus, @{$seqs_aref}[$elem_count-1], $elem, $i_cutoff_cons, $i_cutoff_mfe);
##method

#######################################################################
# Returns the consensus structure of an alignment
# RNAfold
# INPUT:
#	aln_fasta_file		alignment in fasta format
#	cons_file			if parameter provided,
#							extract consensus strcuture from the .cons file in the RFAM given format
# OUTPUT: a reference to the newly created DotPlot object
#######################################################################

sub return_consensus_struct {
	my ( $aln_fasta_file, $cons_file ) = @_;
	my $num_subopt_struct = 0;
	my $consensus;
	if ($cons_file)    #use .cons  file for secondary structure
	{
		$consensus = consensus_file_to_dotbracket($cons_file);
	}
	else               #use PETfold to predict consensus reliable strcuture
	{
		## parse consensus structure file and get it in dot-bracket notation
		die "could not find the alignmnet fasta file $aln_fasta_file!\n"
		  unless ( -e $aln_fasta_file );

		$consensus =
		  call_PETfold_get_cons_strcuture( $aln_fasta_file, $i_threshold_reli, $num_subopt_struct,
			$tmpdir, "$i_job_id.output.pet.cons", $i_job_id );
	}
	if ( !defined $consensus ) {
		print STDERR "ERROR: return_consensus_struct: consensus couldn't be calculted/retrieved\n";
	}
	else {
		print STDOUT "consensus is:\n$consensus\n" if ($i_debug);
	}

	return $consensus;
}

#extract consensus sequence and secondary structure of an stockholm file and write it in a .cons file
sub stockholm_to_cons {
	my ( $stk, $cons_file ) = @_;
	print STDOUT "stockholm_to_cons($stk,$cons_file)\n";

	#	my $cons_file = abs_path($stk.$CONSENSUS_SUFFIX);#append .cons
	my $only_GC = 0;
	my ( $seqs_href, $markup_href ) = MultipleAlignment::Tools::read_stockholm( $stk, $only_GC );

	#	print "keys: " . $_ . "******\n" foreach (keys %{$markup_href});

	die "ERROR: GC markup not found in stockholm file $stk\n" if ( !exists( ${$markup_href}{GC} ) );
	if ( !exists( ${$markup_href}{"GC"}{"RF"} ) ) {
		print STDERR "WARNING: GC RF markup not found in stockholm file $stk\n";
	}
	die "ERROR: GC SS_cons markup not found in stockholm file $stk\n"
	  if ( !exists( ${$markup_href}{"GC"}{"SS_cons"} ) );
	my @cons_elements;
	push( @cons_elements, ">" . ${$markup_href}{"GF"}{"AC"} . "\n" );
	push( @cons_elements, ${$markup_href}{"GC"}{"RF"} . "\n" );
	push( @cons_elements, ${$markup_href}{"GC"}{"SS_cons"} . "\n" );
	print "@cons_elements \n";
	StructureLibrary::Tools::write_file_from_array( $cons_file, \@cons_elements );
	return $cons_file;
}

#
#extract ALIGNED SEQUENCES of an stockholm file and write it in a .aln.fasta file
sub stockholm_to_aln_fasta {
	my ( $stk, $aln_fasta_file ) = @_;
	$aln_fasta_file = abs_path($aln_fasta_file);
	$stk            = abs_path($stk);

	print "stockholm_to_aln_fasta($stk, $aln_fasta_file)\n" if ($i_debug);
	my $cons_file = abs_path( $stk . ".cons" );
	my $only_GC   = 1;
	my ( $seqs_href, $markup_href ) = MultipleAlignment::Tools::read_stockholm( $stk, $only_GC );
	my @fasta_elements;

	#TODO: todo: IMPORTANT seqs order is not preserved!
	foreach my $seq_name ( keys %{$seqs_href} ) {
		push( @fasta_elements, ">" . $seq_name . "\n" );
		push( @fasta_elements, ${$seqs_href}{$seq_name} . "\n" );
	}
	StructureLibrary::Tools::write_file_from_array( $aln_fasta_file, \@fasta_elements );
	return $aln_fasta_file;
}

####
# Reads the consensus structure file and converts the structure to dot-bracket
# format. Dies if anything is wrong with the file or the structure format.
# The consensus structure is in the format of "<>.";
####
sub consensus_file_to_dotbracket {
	my $file = shift;

	my @contents = StructureLibrary::Tools::read_file_return_array($file);

## RFAM families .cons have three lines
#	die "The consensus structure file has the wrong format or could not be read: $file\n" 	unless(@contents == 1) ;

	my $consensus;
	if ( @contents == 1 ) {
		$consensus = $contents[0];
	}
	elsif ( @contents == 3 ) {
		$consensus = $contents[2];
	}
	else {
		die "The consensus structure file has the wrong format or could not be read: $file\n";
	}
	## Remove pseudoknot basepairs, generates warning in stderr
	$consensus = remove_Pknot_basepairs($consensus);

	## Convert infernal/WUSS Secondary Structure Full Format into simple one
	## e.g. :((,<<___>>,,,<<__>>-,))- cpnverted into .((.((...))...((..))..)).
	$consensus =~ tr/<>/()/;
	$consensus =~ tr/{}/()/;
	$consensus =~ tr/[]/()/;
	$consensus =~ tr /_\-,:.~/./;

	if ( $consensus =~ /[()\.]+$/ ) {
		return $consensus;
	}
	else {
		die "The consensus structure \'$consensus\' is in an unknown format in $file!\n";
	}
	-1;
}

####
# This method removes pseudoknot base pairs as defined by stockholm format,
#  i.e. substitute basepairs shown by characters from english alphabet with '.'
sub remove_Pknot_basepairs {
	my $structure = shift;
	print "remove_Pknot_basepairs:\n" . $structure . "\n" if ($i_debug);
	if ( $structure =~ /[a-zA-Z]/ ) {
		print STDERR "WARNING: structure contains pseudoknot basepairs that will be ignored\n";
	}
	$structure =~ s/([a-zA-Z])/\./g;
	print "knotfree:\n" . $structure . "\n" if ($i_debug);
	return $structure;
}

#######################################################################
# Returns the consensus structure of an alignmnet predicted by PETfold
# if num_suboptimals is given and larger than zero, then PETfold is invoked with -r option
#  the optimal structure + the mentioned number of representative strcuture will be returned
# INPUT:
# OUTPUT:
#######################################################################
sub call_PETfold_get_cons_strcuture {
	my (
		$aln_fasta_file, $rel_threshold,    $num_suboptimals,
		$pet_work_dir,   $result_file_name, $job_id
	) = @_;
	my $result_file_full_path = "$pet_work_dir/$result_file_name";
	my $params = "-align-fasta $aln_fasta_file -consensus-out $result_file_full_path ";
	$params .= "-local-config $i_local_config " if $i_local_config;
	$params .= "-threshold $rel_threshold -working-dir $pet_work_dir -job-id $job_id ";
	if ( $num_suboptimals > 0 ) {
		print STDOUT "WARNING: subotmilas not supported in call_PETfold_get_cons_strcuture\n";

		#		$params .= " -r 10000 ";
	}

	#todo: change system for perl files and use current perl interperter system->''
	my $run_PETfold_cmd = "perl $RUN_PETFOLD $params";
	system($run_PETfold_cmd) == 0
	  or die "run_petfold failed code: $?:\n\t $run_PETfold_cmd";
	my $parsed_struct = MultipleAlignment::PETfold::parse_petfold_cons_file($result_file_full_path);
	return $parsed_struct;
}
