#!/usr/bin/env python

import argparse
import os
import sys
import tarfile
import tempfile
import shutil
import errno
import gzip


def is_valid_path(arg):
    if not os.path.exists(arg):
        parser.error("Given path %s does not exist" % arg)
    return arg


def is_no_existing_path(arg):
    if os.path.exists(arg):
        parser.error("Given path %s does exist. Please specify a non existing output file name." % arg)
    return arg


def is_valid_dir(arg):
    arg = is_valid_path(arg)
    if not os.path.isdir(arg):
        parser.error("Given path %s is no directory." % arg)
    return arg


def is_valid_filename(arg):
    arg = is_valid_path(arg)
    if not os.path.isfile(arg):
        parser.error("Given path %s is no file." % arg)
    return arg


def generate_pdf_include(pdf_extraction_path):
    return '\includegraphics[width=\\textwidth]{%s}\n\n' % pdf_extraction_path


def generate_title(doc_title):
    doc_title_clean = doc_title.replace('_', '{\\textunderscore}')
    return '\\title{%s}\n\n' % doc_title_clean


def generate_section(sec_title):
    sec_title_clean = sec_title.replace('_', '{\\textunderscore}')
    return '\section{%s}' % sec_title_clean


parser = argparse.ArgumentParser(description=
                                 """Aggregates plots produced after a certain clustering in one .tex file.
                                    The pdf compiled from this file is written to the given output file name.""")
parser.add_argument('tar_file', type=is_valid_filename,
                    help='The .tar.gz file containing the clustering results and plots to be aggregated.')
parser.add_argument('output_file', type=is_no_existing_path,
                    help='Name of the produced pdf file, e.g., my-results/example.pdf.')
args = parser.parse_args()
tar_file = args.tar_file
output_file = args.output_file

tar_file_base_name = os.path.basename(tar_file)
data_set_name = tar_file_base_name.rstrip('.tar.gz')

# get protein name and binding domains from ENSEMBLE gene ID if this makes sense for the data set at hand
ensemble_gene_id = ''
protein_name = ''
binding_domains = ''
if data_set_name.lower().startswith('ens'):
    ensemble_gene_id = data_set_name.split('_')[0]
    # read file mapping each ensemble gene id to its protein name and RNA binding domain(s)
    ens_to_domains_protein = {}
    with gzip.open('/home/projects/rth_roi/data/RBP/rbp_domain.tsv.gz', 'r') as fin:
        for line in fin:
            (ens, domains, protein) = line.strip().split('\t')
            ens_to_domains_protein[ens] = (domains, protein)
    binding_domains, protein_name = ens_to_domains_protein[ensemble_gene_id]

target_file_names = ['mds-two-dim.pdf',
                     'mds-two-dim-overlap-3D.pdf',
                     'similarity-density.pdf',
                     'hclust-single-result.pdf',
                     'hclust-average-result.pdf',
                     'hclust-complete-result.pdf',
                     'hclust-single-highlighted-overlapping-result.pdf',
                     'hclust-average-highlighted-overlapping-result.pdf',
                     'hclust-complete-highlighted-overlapping-result.pdf']

tmp_dir = None
try:
    # temporary directory used for extracting files
    tmp_dir = tempfile.mkdtemp()
    sys.stderr.write('Created temporary directory %s.\n' % tmp_dir)
    # temporary tex file
    tmp_tex_file = 'tmp.tex'
    tmp_pdf_file = 'tmp.pdf'
    out_file_handle = open(os.path.join(tmp_dir, tmp_tex_file), 'w')
    # the names of each extracted file
    extract_file_list = []
    with tarfile.open(tar_file, 'r:gz') as tar_file_handle:
        for tar_member in tar_file_handle.getmembers():
            for curr_file_name in target_file_names:
                tar_member_filename = str(tar_member.name)
                if tar_member_filename.endswith(curr_file_name):
                    extract_file_list.append(tar_member)

        tar_file_handle.extractall(tmp_dir, extract_file_list)

    latex_preamble = """\documentclass[11pt,notitlepage,a4paper]{article}\n\n\usepackage{graphicx}\n\n\date{}\n\n"""
    latex_intro = """\\begin{document}\n\n\maketitle\n\n"""
    latex_outro = """\end{document}\n"""

    out_file_handle.write(latex_preamble)
    out_file_handle.write(generate_title(data_set_name))
    out_file_handle.write(latex_intro)

    if len(ensemble_gene_id) > 0:
        out_file_handle.write('\\begin{itemize}\n')
        out_file_handle.write('\item %s\n' % ensemble_gene_id)
        out_file_handle.write('\item %s\n' % protein_name)
        out_file_handle.write('\item %s\n' % binding_domains)
        if 'inaccessibility' in data_set_name:
            out_file_handle.write('\item %s\n' % 'only inaccessible binding sites')
        elif 'accessibility' in data_set_name:
            out_file_handle.write('\item %s\n' % 'only accessible binding sites')
        else:
            out_file_handle.write('\item %s\n' % 'binding sites not filtered for accessibility')
        out_file_handle.write('\end{itemize}\n\n')

    # TODO add section titles to add more sturcture to document
    # Set section title(invisilble?) and make sure that shown in header

    # include each plot as a figure in LaTeX document
    for idx in range(len(extract_file_list)):
        extract_file_name = str(extract_file_list[idx].name)
        extract_file_dirname = os.path.dirname(extract_file_name)
        extract_file_dirname = extract_file_dirname[2:]  # cuts off ./
        extract_file_basename = os.path.basename(extract_file_name)

        full_extraction_path = os.path.join(extract_file_dirname, extract_file_basename)
        out_file_handle.write(generate_pdf_include(full_extraction_path))

    out_file_handle.write(latex_outro)
    out_file_handle.close()

    # compile .tex file and copy output pdf to output location specified by the user
    wdir = os.getcwd()
    os.chdir(tmp_dir)
    os.popen('pdflatex %s' % tmp_tex_file).read()
    os.chdir(wdir)
    shutil.copyfile(os.path.join(tmp_dir, tmp_pdf_file), output_file)
finally:
    try:
        shutil.rmtree(tmp_dir)  # delete directory
    except OSError as exc:
        if exc.errno != errno.ENOENT:  # ENOENT - no such file or directory
            raise  # re-raise exception