__author__ = 'Alexander Junge <alexander.junge@gmail.com>'

import os
import sys
import argparse

from Bio import SeqIO


def is_supported_sequence_file_format(arg):
    if arg.lower() == 'stockholm':
        return 'stockholm'
    if arg.lower() == 'fasta':
        return 'fasta'
    else:
        parser.error("Given alignment file type is not supported." % arg)


def is_valid_path(arg):
    if not os.path.exists(arg):
        parser.error("Given path %s does not exist" % arg)
    return arg


parser = argparse.ArgumentParser()
argparse.ArgumentParser(description=
                        '''Computes the average sequence length for all sequences present in a given FASTA/Stockholm
                         file. The sequences are read from stdin and the average sequence length is written to
                         stdout.
                         Alternatively, a list of files can be specified on the command line. In this case,
                         all sequences contained in the given files are pooled and their average sequence length is
                         computed. Furthermore, sequences given on stdin are ignored and average sequence length is
                         still written stdout.''')
parser.add_argument('sequence_format', type=is_supported_sequence_file_format,
                    help='Format of the files storing the sequences to be read, either \'fasta\' or \'stockholm\'')
parser.add_argument('-f', '--filenames', type=is_valid_path, nargs='+', dest='alignment_files',
                    help='File name(s) that contain the sequences to be read. File names in this list must be unique.')
args = parser.parse_args()
SEQUENCE_FORMAT = args.sequence_format
ALIGNMENT_FILES = args.alignment_files

if ALIGNMENT_FILES and len(ALIGNMENT_FILES) > 0:
    # sequences are read from files
    if len(set(ALIGNMENT_FILES)) != len(ALIGNMENT_FILES):
        raise IOError('Names of files given after -f/--filenames must be unique.')

    # collect all sequences read from input files
    sequence_lengths = []
    for input_file in ALIGNMENT_FILES:
        with open(input_file, 'r') as input_file_handler:
            for record in SeqIO.parse(input_file_handler, SEQUENCE_FORMAT):
                # Convert all nucleotides to upper case
                sequence_lengths.append(len(str(record.seq.ungap('-'))))
else:
    # sequences are read from stdin
    input_file_handler = sys.stdin

    sequence_lengths = []
    for record in SeqIO.parse(input_file_handler, SEQUENCE_FORMAT):
        # Remove gaps before computing length
        sequence_lengths.append(len(str(record.seq.ungap('-'))))

# write average sequence length to stdout
if len(sequence_lengths) > 0:
    average_sequence_length = sum(sequence_lengths)/float(len(sequence_lengths))
else:
    average_sequence_length = 0
sys.stdout.write(str(average_sequence_length) + '\n')