#!/usr/bin/env python
import argparse
from Bio import SeqIO
import collections
import logging
import os


logging.basicConfig(level=logging.INFO)

def is_existing_dir(arg):
    if not os.path.isdir(arg):
        parser.error('Given path %s does not point to an existing directory. Exiting.' % arg)
    return arg


def is_not_existing_path(arg):
    if os.path.exists(arg):
        parser.error('Given path %s points to an existing path. Exiting.')
    return arg


__author__ = 'Alexander Junge (alexander.junge@gmail.com)'

parser = argparse.ArgumentParser(description='''
Aggregate sequences for each RNAscClust benchmark set in a single FASTA file.

Takes as input a directory containing all RNAscClust data sets and
collects the sequences in each data set in a single FASTA file called 'all_sequences.fa'.
Furthermore, creates a file that contains all identifiers of representative sequences in the data set called
'representative_ids.txt'.
''')
parser.add_argument('extract_dir', type=is_existing_dir, help='''
Directory containing RNAscClust benchmark data sets as downloaded from Bioinfo Freiburg's FTP server, i.e.,
$ wget ftp://biftp.informatik.uni-freiburg.de/pub/RNAscClust/20151215_data_benchmark_rnascclust.tar.gz
$ tar xzf 20151215_data_benchmark_rnascclust.tar.gz

# the directory required here as positional argument is the data/ directory created by the above extraction.
''')
parser.add_argument('--non_rep_seq_max_count', type=int, default=3,
                    help='number of non-representative sequences from each alignment added to the aggregated data set')
args = parser.parse_args()

extract_dir = args.extract_dir
non_rep_seq_max_count = args.non_rep_seq_max_count

data_set_subdirs = [os.path.join(extract_dir, x)
                   for x in os.listdir(extract_dir) if os.path.isdir(os.path.join(extract_dir, x))]
for subdir in data_set_subdirs:
    non_rep_sequences_filtered = 0
    alignment_dir = os.path.join(subdir, 'alignments')
    stk_files = [os.path.join(alignment_dir, x)
                 for x in os.listdir(alignment_dir)
                 if os.path.isfile(os.path.join(alignment_dir, x)) and x.endswith('.stk')]
    logging.info('Found {:d} .stk files in directory {}.'.format(len(stk_files), alignment_dir))

    representatives_fasta_file = os.path.join(subdir, 'representatives', 'representatives.fa')
    rep_ids = set()
    rep_families = set()
    with open(representatives_fasta_file, 'r') as input_file_handler:
        for record in SeqIO.parse(input_file_handler, 'fasta'):
            seq_id = record.id  # RF00156.3_AC016708.11/153795-153929
            fam_acc = seq_id.split('.')[0]
            rep_id = fam_acc + '_' + seq_id.split('_')[1]  # 'RF00156_AC016708.11/153795-153929'
            rep_ids.add(rep_id)
            rep_families.add(fam_acc)  # RF00156
    logging.info('Found identifiers of {:d} human representative sequences.'.format(len(rep_ids)))
    logging.debug('These are: {}'.format('; '.join(sorted(rep_ids))))
    logging.info('Found {:d} families in human representative sequences.'.format(len(rep_families)))
    logging.debug('These are: {}'.format('; '.join(sorted(rep_families))))

    sequence_list = []
    family_to_seq_count = collections.defaultdict(int)
    rep_ids_extended = set()
    for stockholm_file in sorted(stk_files):
        curr_non_rep_seq_count = 0
        object_id = os.path.basename(stockholm_file).rstrip('.stk')
        fam_acc = object_id.split('.')[0]
        if fam_acc not in rep_families:
            continue
        with open(stockholm_file, 'r') as input_file_handler:
            for record in SeqIO.parse(input_file_handler, 'stockholm'):
                seq_count = family_to_seq_count[fam_acc]
                # append sequence index to identifier. object_id_extended will then look like this: RF00005.66, which
                # refers to the 67th sequence beloging to family RF00005
                object_id_extended = fam_acc + '.' + str(seq_count)
                family_to_seq_count[fam_acc] += 1
                new_record_id = object_id_extended + '_' + record.id

                is_rep_sequence = False
                if fam_acc + '_' + record.id in rep_ids:
                    rep_ids_extended.add(object_id_extended)
                    is_rep_sequence = True
                else:
                    curr_non_rep_seq_count += 1

                if is_rep_sequence or curr_non_rep_seq_count <= non_rep_seq_max_count:
                    record.seq = record.seq.ungap('-')
                    record.id = new_record_id
                    record.description = object_id_extended
                    sequence_list.append(record)
                else:
                    non_rep_sequences_filtered += 1
    logging.info('Filtered {:d} non-representative sequences.'.format(non_rep_sequences_filtered))

    assert len(rep_ids_extended) == len(rep_ids)

    with open(os.path.join(subdir, 'all_sequences.fa'), 'w') as output_handle:
        SeqIO.write(sequence_list, output_handle, 'fasta')

    with open(os.path.join(subdir, 'representative_ids.txt'), 'w') as output_handle:
        for rep_id in sorted(rep_ids_extended):
            output_handle.write(rep_id + os.linesep)

    readme_content = '''
Contains one sub-directory per data set:

- rfamome
- cliques-high
- cliques-medium
- cliques-low

Each of these sub-directory furthermore contains the following directories and files:

- alignments/ - dirctory containing the alignments to be clustered in Stockholm format
- all_sequences.fa - sequences contained in all alignments in directory alignments/ aggregated in a single FASTA file
- representatives/ - directory containing a single FASTA file containing the representative sequences in each alignment
- representative_ids.txt - lists the sequence identifiers of representative sequences used in the all_sequences.fa file.
                           This file is needed to identify representative sequences in all_sequences.fa.
'''
    with open(os.path.join(extract_dir, 'README'), 'w') as fout:
        fout.write(readme_content)
