#!/usr/bin/env python

__author__ = 'Alexander Junge (alexander.junge@gmail.com)'

import argparse
import os
import gzip
import collections
import re
import sys
import logging


def is_positive_integer(arg):
    value = None
    try:
        value = int(arg)
    except ValueError:
        parser.error("Given length threshold must be an integer but is %s." % arg)
    if value < 0:
        parser.error("Given length threshold must be positive but is %d." % value)
    return value


def is_valid_path(arg):
    if not os.path.exists(arg):
        parser.error("Given path %s does not exist. Please specify an existing file." % arg)
    return arg


def is_not_valid_path(arg):
    if os.path.exists(arg):
        parser.error("Given path %s does exist. Please specify a non-existing path.")
    return arg


logging.basicConfig(level=logging.INFO)
parser = argparse.ArgumentParser(description=
                                 """
                                 Filters Rfam families with too few instances from a GSPAN file and writes remainder
                                 to stdout.
                                 """)
parser.add_argument('GSPAN', type=is_valid_path,
                    help="""
                         The input GSPAN file to be filtered.
                         """)
parser.add_argument('-n', '--mincount', default=3, type=is_positive_integer,
                    help="""
                        Filtering threshold. Rfam Families with less members that the given number are removed.
                        """)
parser.add_argument('-g', '--GraphClust', action='store_true',
                    help="""
                         Set this flag if the input GSPAN file was generated by GraphClust.
                         If this flag is not set, the GSPAN file expected to be generated by RNAalignClust.
                         """)
parser.add_argument('--filtered-format', action='store_true',
                    help="""
                         Assuming RNAalignClust format with SEQID of for RF00001_filtered.0.folded
                         """)
parser.add_argument('-o', '--oldRNAalignClust', action='store_true',
                    help="""
                         Set this flag if the old RNAalignClust GSPAN format is to be parsed.
                         """)
parser.add_argument('-s', '--write-sequence-ids', type=is_not_valid_path,
                    help="""
                         If a filename is given here, the sequence identifiers in the data set will be written to it.
                         """)

args = parser.parse_args()

if args.GraphClust and args.oldRNAalignClust:
    parser.error('-f and -o option may not be set simultaneously.')

family_to_instances = collections.defaultdict(set)
if args.GraphClust:
    logging.info('Assuming GSPAN is in GraphClust format.')
    seqid_regex = re.compile('t #.+ORIGHEAD\s(RF\d+_rep\.\d+)')
else:
    logging.info('Assuming GSPAN is in RNAalignClust format.')
    if args.filtered_format:
        logging.info('Assuming RNAalignClust format with SEQID of form RF00001_filtered.0.folded')
        seqid_regex = re.compile('\w\s#\sSEQID\s(RF\d+\_filtered.\d+.folded)')
    
    elif args.oldRNAalignClust:
       logging.info('Assuming GSPAN is in old RNAalignClust format.')
       seqid_regex = re.compile('\w\s#\sSEQID\s(RF\d+\.\d+)')
    else:
       logging.info('Assuming GSPAN is in new RNAalignClust format.')
       seqid_regex = re.compile('t\s#\sSEQID\s(RF\d+\.\d+)')

with gzip.open(args.GSPAN, 'r') as gspan_file:
    for line in gspan_file:
        line_match = seqid_regex.match(line)
        if line_match:
            seq_id = line_match.group(1)
            if args.filtered_format:
                seq_id = seq_id.replace("_filtered", "").replace(".folded", "")
            fam_id = seq_id.split('.')[0]
            family_to_instances[fam_id].add(seq_id)

sys.stderr.write('Found {:d} Rfam families:\n{}\n\n'.format(len(family_to_instances),
                                                            ','.join(family_to_instances.keys())))

valid_families = [fam_id
                  for fam_id, instance_set in family_to_instances.iteritems() if len(instance_set) >= args.mincount]
sys.stderr.write('Found {:d} Rfam families that have at least {:d} instances:\n{}\n\n'.format(len(valid_families),
                                                                                              args.mincount,
                                                                                              ','.join(sorted(valid_families))))
print_curr_line = False
out_seqids = []
with gzip.open(args.GSPAN, 'r') as gspan_file:
    for line in gspan_file:
        line_match = seqid_regex.match(line)
        if line_match:
            seq_id = line_match.group(1)
            if args.filtered_format:
                seq_id = seq_id.replace("_filtered", "").replace(".folded", "")
                # remove filtered and folded tags of output gspan to have compatibility from now on
                line = line.replace("_filtered", "").replace(".folded", "") 
            fam_id = seq_id.split('.')[0]
            if fam_id in valid_families:
                out_seqids.append(seq_id)
                print_curr_line = True
            else:
                print_curr_line = False
        if print_curr_line:
            sys.stdout.write(line)

if args.write_sequence_ids:
    with open(args.write_sequence_ids, 'w') as seqout:
        for sequence_id in out_seqids:
            seqout.write(sequence_id)
            seqout.write('\n')