#!/usr/bin/env python

__author__ = 'Alexander Junge (alexander.junge@gmail.com)'

import logging
import collections

class EDeNClusterLoader:
    """
    Wrapper loading pre-computed clustering results computed by the C++ EDeN implementation once the method
    fit_predict is called.
    """

    def __init__(self, cluster_file_path, target_file_path, use_majority_cluster=False, object_id_file_path=None):
        """

        :param cluster_file_path: location of the file generated by C++ EDeN that contains the clustering result
        :param target_file_path: location of the file generated by C++ EDeN that contains the target clustering
        :param use_majority_cluster: boolean, if this is set, each object is allowed to be assigned to multiple clusters
                                            and fit_predict automatically assigns the object to the cluster most of its
                                            'copies' are assigned to, the so-called majority cluster. If there is more
                                            than one majority cluster,a random majority cluster is chosen.
                                            If this is not set, assigning an object to multiple clusters will throw an
                                            IOError.
        :param object_id_file_path: location of the file holding the object identifiers. This file must only be given if
                                    use_majority_cluster is True and contained object identifiers must be in the same
                                    order as in the file target_file_path.
        """
        self.cluster_path = cluster_file_path
        self.target_path = target_file_path
        self.use_majority_cluster = use_majority_cluster
        if use_majority_cluster and not object_id_file_path:
            raise ValueError('object_id_file_path must be given if use_majority_cluster is set.')
        self.object_id_path = object_id_file_path

    def __load_standard_clustering__(self):
        object_2_cluster = {}
        with open(self.cluster_path, 'r') as cluster_file:
            for cluster_number, line in enumerate(cluster_file):
                for obj in line.split():
                    if obj in object_2_cluster:
                        raise IOError('Input C++ EDeN clustering result read from {} assigns object {} to '
                                      'more than one cluster.'.format(self.cluster_path, obj))
                    else:
                        object_2_cluster[obj] = cluster_number
        logging.debug('Loaded cluster assignments for {:d} objects.'.format(len(object_2_cluster)))
        return object_2_cluster

    def __load_object_identifiers__(self):
        obj_ids = []
        with open(self.object_id_path, 'r') as object_file:
            for line in object_file:
                obj_id = line.rstrip('\r\n')
                obj_ids.append(obj_id)
        return obj_ids

    def __load_majority__clustering(self):
        object_identifier = self.__load_object_identifiers__()
        object_2_clusters = collections.defaultdict(list)
        with open(self.cluster_path, 'r') as cluster_file:
            for cluster_number, line in enumerate(cluster_file):
                for obj_number in line.split():
                    obj = object_identifier[obj_number]
                    object_2_clusters[obj].append(cluster_number)

        # assign majority cluster, if more than one majority cluster exists, assign random one
        object_to_majority_cluster = {}
        for obj, cluster_ids in object_2_clusters.iteritems():
            majority_cluster = collections.Counter(cluster_ids).most_common(1)[0][0]
            object_to_majority_cluster[obj] = majority_cluster

        logging.debug('Loaded cluster assignments for {:d} objects.'.format(len(object_2_clusters)))
        return object_to_majority_cluster

    def __load_clustering__(self):
        if self.use_majority_cluster:
            return self.__load_majority__clustering()
        else:
            return self.__load_standard_clustering__()

    def fit_predict(self, matrix):
        """
        Loads and returns a pre-computed C++ EDeN clustering result.

        :param matrix: a (dis)similarity matrix. Note that this matrix is not used as only a previously computed
                       clustering is loaded here.
        :return: list of strings, encoding the PREDICTED clustering result. The cluster assignment of object at
                 index i is found at position i of this list.
        """
        if matrix:
            logging.warning('Given (dis)similarity matrix is ignored since a C++ EDeN clustering is loaded.')

        object_2_cluster = self.__load_clustering__()
        # TODO adapt code below to mulitple instances/object issue
        obj_count = len(object_2_cluster)
        all_objects = range(obj_count)
        cluster_assignment = []
        for curr_obj in all_objects:
            curr_obj = str(curr_obj)
            if not curr_obj in object_2_cluster:
                raise IOError('Input C++ EDeN clustering result read from {} does not assign object {} '
                              'to a cluster.'.format(self.cluster_path, curr_obj))
            else:
                cluster_assignment.append(object_2_cluster[curr_obj])

        # sanity check that number of objects is the same as in the cluster result and target clustering
        if len(self.load_eden_target_values()) != len(cluster_assignment):
            raise IOError('Loaded C++ clustering and EDeN target clustering contain different numbers of objects.')

        return cluster_assignment

    def load_eden_target_values(self):
        """

        :return: list of strings, encoding the GROUND TRUTH clustering result. The target cluster assignment of object
                 at index i is found at position i of this list.
        """
        #TODO modify in order to deal with mulitple instances/object issue
        target_values = []
        with open(self.target_path, 'r') as target_file:
            for line in target_file:
                target = line.strip()
                target_values.append(target)
        return target_values