import numpy as np
from scipy.sparse import csr_matrix
import itertools

# parameters setting
CHUNK_SIZE = 100



def exclude_single_partitions(feat_arr, targets, min_num_parts=2):
    # count num inst per target
    from collections import Counter
    counts =  Counter(targets)
    from scipy.sparse import vstack
    assert(feat_arr.shape[0] == len(targets))
    nonsingles_arr = None
    for i, target in enumerate(targets):
        if counts[target] >= min_num_parts:
            if nonsingles_arr is None:
                nonsingles_arr = feat_arr[i, :]
            else:
                nonsingles_arr = vstack([nonsingles_arr, feat_arr[i]], format="csr")
    nonsingles_targets = [t for t in targets if (counts[t] >= min_num_parts)]

    return nonsingles_arr, nonsingles_targets



def save_sparse_csr(filename, array):  # TODO: use savez_compress
    import numpy as np
    np.savez(filename, data=array.data, indices=array.indices,
             indptr=array.indptr, shape=array.shape)
    

def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    from itertools import izip_longest
    return izip_longest(*args, fillvalue=fillvalue)


def print_vect_info(V):
    print 'Instances: %d Features: %d with an avg of %d features per instance' \
        % (V.shape[0], V.shape[1],  V.getnnz()/V.shape[0])


def consume_iterator(iterator, n):
    "Advance the iterator n-steps ahead. (from itertools recipes)"
    # Use functions that consume iterators at C speed.
    # advance to the empty slice starting at position n
    next(itertools.islice(iterator, n, n), None)


def minimal_gspan_file(input_file, output_file):
    ''' Converts gspan into minimal format
     keeps the first three and four entries for vertices and edges respectively'''
    # bash alternative:
    # cat file.gspan | \
    # awk'{if($1=="v"||$1=="V"){print $1,$2,$3}else if($1=="e"){print $1,$2,$3,$4}else{print $0}}'\
    # > file.gspan2
    import re
    with open(input_file, 'r') as in_handle, open(output_file, 'w') as out_handle:
        for line in in_handle:
            # discard stacking and abstract graph entries, # TODO: Support them!
            if re.search(r' n$| p$| P$|AS#|@|\^', line):
                continue
            line = line.rstrip()
            if line.startswith("V") or line.startswith("v"):
                cols = line.split()
                out_line = " ".join(cols[0:3])
            elif line.startswith("e"):
                cols = line.split()
                out_line = " ".join(cols[0:4])
            else:
                out_line = line
            out_handle.write(out_line+"\n")


def load_sparse_csr(filename):
    loader = np.load(filename)
    return csr_matrix((loader['data'], loader['indices'], loader['indptr']), shape=loader['shape'])

 
def parse_gspan_header(gspan, graphclust_format=False):
    """ Extract annotation reference from GSPAN instance header """
    import re
    instances = []
    classes = {}
    with open(gspan, 'r') as f:
        for line in f:
            if line.startswith('t #'):  # start of a new instance, line contains tags,etc
                tokens = line.split()
                match_class = ""
                match_subclass = ""
                if (not graphclust_format):  # alignclust gspan header format
                    tag = tokens[3]   # tag is the column containing instance id
        #             print tag
    #                 matches = re.match(r"random_subalign_all_5_10_(RF.*)\.(\S+)\.folded",tag)
                    matches = re.match(r"^(RF.*)\.(\S+)\.folded", tag)
                    if (not matches):
                        raise IOError('Cannot find id info from instance header:' + line)
                    match_class = matches.group(1)
                    match_subclass = matches.group(2)
                else:  # GraphClust gspan tag
                    tag = tokens[8]  # tag is the column containing instance id
    #                 print tag
                    matches = re.match(r"^(RF[0-9]*)_rep\.(\S+)", tag)
                    if (not matches):
                        raise IOError('Cannot find id info from instance header:' + line)
                    match_class = matches.group(1)
                    match_subclass = matches.group(2)
                if(match_class not in classes):
                    classes[match_class] = len(classes)+1
                class_id = classes[match_class]

                instances.append(dict([('class',       match_class),
                                       ('subclass',    match_subclass),
                                       ('instance_id', len(instances)+1),
                                       ('class_id',    class_id)]))

    return instances
   # end parse_gspan_header

# =========================


class GSPANVectorizer:
    def __init__(self, gspan_file, out_dir, saved_npz_vector=None, graphclust_format=False,
                 num_jobs=-1, max_num_instances=-1, max_radius=2, min_radius=0, max_distance=2,
                 simplify_gspan=True):

        # ===checks===
        import os.path
        if not os.path.isfile(gspan_file):
            raise IOError('GSPAN file does NOT exist. [%s]' % gspan_file)

        self.out_dir = out_dir + "/"
        if not os.path.isdir(self.out_dir):
            raise IOError('out_dir does NOT exist. [%s] ' % self.out_dir)
        # TODO: read tar.gz files

        # ===Set self attributes===
        self.num_jobs = num_jobs
        self.graphclust_format = graphclust_format
        # limit number of features vectorized from gspan file
        self.MAX_VECTORIZED_FEATURES = max_num_instances

        # ===convert gspan file to minimal format===
        if simplify_gspan:
            self.gspan = gspan_file + "_minimal"
            minimal_gspan_file(gspan_file, self.gspan)
        else:
            self.gspan = gspan_file

        from eden.converter.graph import gspan
        # limit number of features vectorized from gspan file
        graphs = gspan.gspan_to_eden(input=self.gspan)  # , input_type='file')
        if max_num_instances >= 0:
            self.eden_graphs = itertools.islice(graphs, max_num_instances)
        else:
            self.eden_graphs = graphs

        self.instances_info = parse_gspan_header(self.gspan, self.graphclust_format)
        self.ref_labels = [x['class'] for x in self.instances_info]
        self.ref_labels_id = [x['class_id'] for x in self.instances_info]
        # load  previously vectorized object from file, vectorization may be partial
        # load saved array
        self.start_instance = 0
        self.current_npz_file = ""
        self.X_loaded = None
        self.X = None
        if saved_npz_vector:
            if not os.path.isfile(saved_npz_vector):
                print ('ERROR: NPZ vector file does NOT exist. [%s]' % saved_npz_vector)
            else:
                self.X_loaded = load_sparse_csr(saved_npz_vector)
                # skip loaded graphs
                consume_iterator(self.eden_graphs, self.X_loaded.shape[0])
                self.current_npz_file = saved_npz_vector
                print "Info: {} instances loaded from npz file" .format(self.X_loaded.shape[0])

        # load pyEDeN graph vectorizer
        from eden import graph
        self.eden_vectorizer = graph.Vectorizer(r=max_radius, d=max_distance, min_r=min_radius)

    # end init

    def save_sparse_csr(self, filename, array):  # TODO: use savez_compress
        np.savez(filename, data=array.data, indices=array.indices,
                 indptr=array.indptr, shape=array.shape)
        self.current_npz_file = filename


    def do_vectorize(self):
        from eden.util import vectorize
        from itertools import islice
        self.X = None
        # EDeN vectorizer seems to have memory leakage so we set block_size=1 do chunk/block here
        self.X = vectorize(islice(self.eden_graphs, CHUNK_SIZE), vectorizer=self.eden_vectorizer,
                           block_size=1, n_jobs=self.num_jobs)

        from scipy.sparse import vstack
        if self.X_loaded is not None:  # if npz loaded, prepend npz array to X
            assert(self.X.shape[0] == 0 or
                   self.X.shape[1] == self.X_loaded.shape[1])  # number of features must match

            self.X = vstack([self.X_loaded, self.X], format="csr")  # prepend loaded npz array to X

        print ('1st: Instances: %d Features: %d with an avg of %d features per instance'
               % (self.X.shape[0], self.X.shape[1], self.X.getnnz()/self.X.shape[0]))
        self.save_sparse_csr(self.out_dir + "/X_array_" + str(self.X.shape[0]) + ".npz", self.X)

        for chunk in grouper(self.eden_graphs, CHUNK_SIZE):
            X_chunk = vectorize(chunk, vectorizer=self.eden_vectorizer,
                                block_size=1, n_jobs=self.num_jobs)
            self.X = vstack([self.X, X_chunk], format="csr")
            print ('Chunk Instances: %d Features: %d with an avg of %d features per instance'
                   % (X_chunk.shape[0], X_chunk.shape[1],  X_chunk.getnnz()/X_chunk.shape[0]))
            print ('Instances: %d Features: %d with an avg of %d features per instance'
                   % (self.X.shape[0], self.X.shape[1],  self.X.getnnz()/self.X.shape[0]))
            self.save_sparse_csr(self.out_dir + "/X_array_" + str(self.X.shape[0]) + ".npz", self.X)
        return self.current_npz_file
    # end vectorize_gspan

    def get_feature_vector(self):
        # TODO: return one already loaded by vectorize()?
        loaded_arr = load_sparse_csr(self.current_npz_file)
        if (self.MAX_VECTORIZED_FEATURES > 0 and
            self.MAX_VECTORIZED_FEATURES < loaded_arr.shape[0]):
            return loaded_arr[:self.MAX_VECTORIZED_FEATURES]
        else:
            return loaded_arr
