#!/usr/bin/env python
import sys
import os
import argparse

import collections
from sklearn import metrics, cross_validation
from sklearn.neighbors import KNeighborsClassifier
import numpy as np
import logging
from scipy.sparse import csr_matrix
from collections import Counter
import gzip
import mimetypes
import shutil
from sklearn.cross_validation import StratifiedKFold, KFold
from sklearn.utils.validation import _num_samples

# Import libraries located in relation to this file
import inspect
current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parent_dir = os.path.dirname(current_dir)
bin_dir = parent_dir
sys.path.insert(0, bin_dir)
from analysis.tools import mcc_cluster
from analysis.tools.PairCounting import counts as pair_counts

logging.basicConfig(level=logging.INFO)


def is_existing_dir(arg):
    if not os.path.isdir(arg):
        parser.error("Given path %s is not a directory. Exiting." % arg)
    return arg


def knn_cross_validation(x_matrix, targets, kfold=3, dist_metric='cosine', non_stratified_cv=False):
    """
    Fits K-Nearest Neighbor model to given data and returns cross validation error under various performance metrics and
    for different choices of K.

    :param x_matrix: data to fit the KNN model to
    :param targets: the target variable to predict
    :param kfold: the number of folds to be used in (stratified) K-fold cross validation
    :param dist_metric: the distance metric to be used in KNN model
    :param non_stratified_cv: If this is set, standard, non-stratified cross validation is used
    :return: list of tuples of str.
        Model accuracy under various performance metrics achieved by KNN model with different choices of K.

        Performance metrics are: 'precision_weighted', 'recall_weighted', 'f1_weighted'
        The tuple contains: name of the performance measure, choice of K,
        mean and standard deviation of the performance measure over all cross-validation folds.
        Example: [('precision_weighted', '1', '0.81', '0.01'), ('recall_weighted', '1', '0.52', '0.021'),
                  ('f1_weighted', '1', '0.75', '0.022')]
    """
    logging.debug('KNN targets: %s' % ', '.join(targets))
    performance_list = []

    for n in [1, 3, 5, 7, 9]:
        for metric in ['precision_weighted', 'recall_weighted', 'f1_weighted']:
            # Note: One can use weighted vote instead of majority vote with param: weights = 'distance'
            neigh = KNeighborsClassifier(n_neighbors=n, algorithm='brute', metric=dist_metric)
            if non_stratified_cv:
                cv_fun = KFold(_num_samples(targets), kfold, shuffle=True, random_state=42)
            else:
                cv_fun = StratifiedKFold(targets, kfold, shuffle=True, random_state=42)
            logging.captureWarnings(True)
            scores = cross_validation.cross_val_score(neigh, x_matrix, targets, cv=cv_fun, scoring=metric, n_jobs=1)
            logging.captureWarnings(False)
            # noinspection PyStringFormat
            performance_list.append((metric, str(n), str(np.round(np.mean(scores), 2)),
                                     str(np.round(np.std(scores), 2))))
    return performance_list


def clustering_report(x_matrix, targets, dis_matrix, model, predicts=None, one_noise_class=False, minimal=True,
                      write_header=True, latex=False):
    """
    Performs an analysis of a clustering result and returns the clustering performance under various quality metrics.

    :param x_matrix: the feature matrix, only needed for computing the silhouette coefficient
    :param targets: the true cluster assignment of each object in the data set
    :param dis_matrix: the pairwise distance matrix, only needed if parameter model is given
    :param model: an optional clustering model to predict clusters based on dis_matrix parameter
    :param predicts: the predicted clusters; needed if model parameter is not specified
    :param one_noise_class: if True, outlier objects that are not assigned to a cluster are assigned to a single cluster
    :param minimal: if True, a minimal set of clustering performance metrics is computed
    :param write_header: if True, a header will be added to the returned clustering report
    :param latex: if True, the clustering metrics will be written in a TeX friendly format, i.e., & separated
    :return: a str; the clustering result measured under different distance metrics
    """
    if predicts is None:
        labels = model.fit_predict(dis_matrix)
    else:
        labels = list(predicts)
    assert (len(labels) == len(targets))

    logging.debug('Predicted: %s' % ', '.join(labels))
    logging.debug('Targets: %s' % ', '.join(targets))

    # assign unique label to each outlier/unpredicted sample
    next_outlier_class_label = 10000000
    for i in range(len(labels)):
        if one_noise_class:
            if labels[i] == -1 or labels[i] == '-1' or str(labels[i]).startswith('unassigned-'):
                labels[i] = 'unassigned'
        else:
            if labels[i] == -1:
                labels[i] = next_outlier_class_label
                next_outlier_class_label += 1
            elif labels[i] == '-1' or str(labels[i]).startswith('unassigned-'):
                labels[i] = str(next_outlier_class_label)
                next_outlier_class_label += 1

    labels_true = targets
    label_count = len(set(labels))
    object_count = len(labels)

    class_count = len(set(labels_true))
    cluster_count = len(set(labels))

    precision = recall = f_1 = f_2 = -999
    try:
        pcounts = pair_counts(labels_true, labels)
        precision = pcounts['TP'] / float(pcounts['TP'] + pcounts['FP'])
        recall = pcounts['TP'] / float(pcounts['TP'] + pcounts['FN'])
        beta = 1
        f_1 = (beta * beta + 1) * precision * recall / (beta * beta * precision + recall)
        beta = 2
        f_2 = (beta * beta + 1) * precision * recall / (beta * beta * precision + recall)
    except ZeroDivisionError:
        logging.warn("Division by zero by pair counting calculations")

    if 1 < label_count < object_count and x_matrix is not None:
        sil = metrics.silhouette_score(x_matrix, labels)
    else:
        sil = -999.0

    # adjusted mutual information
    ami = metrics.adjusted_mutual_info_score(labels_true, labels)

    # Rand index according to https://en.wikipedia.org/wiki/Rand_index
    n_samples = len(labels_true)
    a = 0
    b = 0
    pair_count = 0
    for i in range(n_samples - 1):
        for j in range(i + 1, n_samples):
            pair_count += 1
            if (labels[i] == labels[j]) and (labels_true[i] == labels_true[j]):
                a += 1
            if (labels[i] != labels[j]) and (labels_true[i] != labels_true[j]):
                b += 1
    ri = (a + b) / float(pair_count)

    scores = [metrics.homogeneity_score(labels_true, labels),
              metrics.completeness_score(labels_true, labels),
              metrics.v_measure_score(labels_true, labels),
              ri,
              metrics.adjusted_rand_score(labels_true, labels),
              mcc_cluster.clustering_matthews_corrcoef(labels_true, labels),
              cluster_count, class_count
              ]

    header = "Homogeneity, Completeness, V-measure, RandIndex, AdjustedRandIndex, Matthew's correlation coefficient," \
             " ClusterCount, TrueClassCount"

    if not minimal:
        scores += [ami, sil, precision, recall, f_1, f_2]
        header += ", AjustedMutualInformation, Silhouette Coefficient, Precision, Recall, f_1, F2"

    rounded_scores = []
    for score in scores:
        rounded_scores.append("%0.3f" % score)

    if latex:
        score_str = os.linesep + "& ".join(rounded_scores)
    else:
        score_str = ", ".join(rounded_scores)
    if not write_header:
        header = ""
    if one_noise_class:
        header = os.linesep + "***One class for all outliers." + os.linesep + header
    return header + os.linesep + score_str


def read_nspdk_feature(feature_file, cutoff=0.01):
    rows = list()
    cols = list()
    vals = list()
    nnz_per_row = list()
    if mimetypes.guess_type(feature_file)[1] == 'gzip':
        fopen = gzip.open
    else:
        fopen = open
    with fopen(feature_file, 'rb') as f:
        nonzeros = 0
        for rownum, line in enumerate(f):
            for pair in line.split():
                pairparts = pair.split(":")
                assert len(pairparts) == 2
                try:
                    colnum = int(pairparts[0])
                    value = float(pairparts[1])
                except ValueError:
                    logging.error("Line:{} cannot parse pair: {}".format(rownum, pairparts))
                    sys.exit(-1)
                rows.append(rownum)
                cols.append(colnum)
                vals.append(value)
                nonzeros += 1
            # end for pair
            nnz_per_row.append(nonzeros)
            # end for line
    sparse_mat = csr_matrix((vals, (rows, cols)))
    logging.info("filtered size is %d" % len([v for v in vals if v > cutoff]))
    logging.info("Shape: {} #non-zeros {} #non-zeros-per-row: {}".format(sparse_mat.shape, sparse_mat.nnz,
                                                                         sum(nnz_per_row) / max(1, len(nnz_per_row))))
    return sparse_mat


def read_fasta_targets(gc_fasta):
    gc_targets = list()
    if mimetypes.guess_type(gc_fasta)[1] == 'gzip':
        fopen = gzip.open
    else:
        fopen = open
    with fopen(gc_fasta, "r") as fasta:
        for line in fasta:
            if line.startswith(">"):
                subfam = line.split()[1]
                fam = subfam[0:7]
                if not fam.startswith("RF"):
                    raise ValueError("Unknown target fam in %s " % subfam)
                gc_targets.append(fam)
    logging.info("#targets: %d" % len(gc_targets))
    return gc_targets


def eval_clustering_result(results_file_path, fasta_file_path, minimal=True):
    """

    :param results_file_path: str
     Path to file containing predicted clusters that is assumed to contain two columns per line. First column
     is predicted cluster (an integer), second is object name.
     Example line: '2<TAB>RF00001.11'
    :param fasta_file_path: str
     Path to FASTA file containing sequences in each class that used to map clustered objects to their true families.
     E.g., file may contain line '> RF00001.1' would result in mapping RF00001.1 to its true family RF00001
    :param minimal: If True, the minimal required set of clustering metrics is computed
    :return: a clustering evaluation as returned by the method clustering_report()
    """
    object_to_true = dict()
    object_to_predicted = dict()
    if mimetypes.guess_type(fasta_file_path)[1] == 'gzip':
        fopen = gzip.open
    else:
        fopen = open
    with fopen(fasta_file_path, "r") as fasta:
        for line in fasta:
            if line.startswith(">"):
                obj = line.split()[1]
                fam = obj[0:7]
                if not fam.startswith("RF"):
                    raise ValueError("Unknown target fam in %s." % obj)

                object_to_true[obj] = fam
                # in order to map unclustered sequences to singleton clusters
                object_to_predicted[obj] = ['unassigned-' + obj]

    if mimetypes.guess_type(results_file_path)[1] == 'gzip':
        fopen = gzip.open
    else:
        fopen = open
    with fopen(results_file_path, "r") as f:
        for line in f:
            cluster_id = line.split()[0]
            obj = line.split()[1]
            if object_to_predicted[obj][0].startswith('unassigned-'):  # obj seen first, create list of pred. clusters
                object_to_predicted[obj] = [cluster_id]
            else:
                object_to_predicted[obj].append(cluster_id)

    # check how many objects are assigned to more than one or no cluster
    for current_object, current_cluster in object_to_predicted.iteritems():
        if len(current_cluster) != 1:
            logging.warning('Object %s is not assigned to exactly one cluster, but to the following clusters:%s%s' %
                            (current_object, os.linesep, ', '.join(current_cluster)))
        if current_cluster[0].startswith('unassigned-'):
            logging.debug('Object %s is not assigned to a cluster.' % current_object)

    all_objects = sorted(object_to_true.keys())
    true_list = [object_to_true[obj] for obj in all_objects]
    # use majority class for predictions
    predicted_list = [Counter(object_to_predicted[obj]).most_common()[0][0] for obj in all_objects]

    all_clusters = sorted(set(predicted_list))
    cluster_to_index = {}
    for i, current_cluster in enumerate(all_clusters):
        cluster_to_index[current_cluster] = i
    logging.info('Found {:d} predicted clusters and {:d} true clusters for {:d} objects.'.format(len(cluster_to_index),
                                                                                                 len(set(true_list)),
                                                                                                 len(all_objects)))

    # a list with one item per cluster; each cluster is itself represented as a list of object identifiers
    cluster_list = [[] for i in range(len(all_clusters))]
    for i, obj in enumerate(all_objects):
        obj_cluster = predicted_list[i]
        cluster_list[cluster_to_index[obj_cluster]].append(obj)

    ret = clustering_report(None, true_list, None, model=None, predicts=predicted_list, minimal=minimal)
    return ret, cluster_list


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='''
                                                 Evaluates a clustering result.

                                                 By default, an RNAscClust result is expected.
                                                 ''')
    parser.add_argument('clustering_result_dir', type=is_existing_dir,
                        help='''
                             Directory containing clustering result to be analyzed. The directory is assumed to have
                             the default structure as created by the clustering programme.
                             ''')
    parser.add_argument('output_dir',
                        help='''
                             Output files and plots are written to this directory.

                             If the directory does not exist it will be created at runtime.
                             Two output files will be created in directory:
                             `knn_analysis.tsv` contains output of KNN analysis;
                             `cluster_analysis.tsv` contains output of clustering analysis
                             ''')
    parser.add_argument('--graph_clust', action='store_true',
                        help='set this, if a GraphClust clustering is to be analyzed instead of an RNAscClust result')
    parser.add_argument('--graph_clust_all_sequences', action='store_true',
                        help='set this, if a GraphClust clustering obtained after runnning the program with all '
                             'sequences (and not only with the representative sequences) is to be evaluated')
    parser.add_argument('--no_knn_analysis', action='store_true',
                        help='set this to avoid KNN analysis step')
    parser.add_argument('--gzipped_feature_vectors', action='store_true',
                        help='If this is set, feature vector files are assumed to be gzipped and hence .gz is appended '
                             'to default file names')
    parser.add_argument('--cvfolds', type=int, default=3,
                        help='the number of Cross-Validation folds to use in K-nearest neighbor classifier')
    parser.add_argument('--no_stratified_cv', action='store_true',
                        help='when this is set, standard, non-stratified Cross-Validation is used in K-nearest neighbor'
                             ' classifier')
    args = parser.parse_args()

    input_dir = args.clustering_result_dir
    output_dir = args.output_dir
    if not os.path.exists(output_dir):
        try:
            os.makedirs(output_dir)
        except OSError:
            pass

    is_graph_clust = args.graph_clust
    graph_clust_all_sequences = args.graph_clust_all_sequences
    if graph_clust_all_sequences:
        is_graph_clust = True

    no_knn_analysis = args.no_knn_analysis
    gzipped_feature_vectors = args.gzipped_feature_vectors
    cvfolds = args.cvfolds
    no_stratified_cv = args.no_stratified_cv

    if is_graph_clust:  # GraphClust
        feature_vector_file = os.path.join(input_dir, 'SVECTOR', 'feat.data.svector.1')
        feature_vector_file_knn = os.path.join(input_dir,'GSPAN', '1.group.gspan_20bits.feature')
            
    else:  # RNAscClust
        feature_vector_file = os.path.join(input_dir, 'encoded_alignments.gspan.feature')
        feature_vector_file_knn = os.path.join(input_dir, 'encoded_alignments_20bits.gspan.feature')

    if gzipped_feature_vectors:
        feature_vector_file += '.gz'

    cluster_result_file = os.path.join(input_dir, 'clustering_results_overall.txt')
    
    if is_graph_clust and not os.path.exists(cluster_result_file):  # Collect overall_results in the case of GraphClust
        cluster_dirs = os.path.join(input_dir, 'RESULTS', '[0-9]*')
        # One liner collecting clustering results
        collector_cmd = "for c in " + cluster_dirs + "; do cat $c/cluster.all "\
                        + "| awk -v c=$(basename $c) '{print c,$11}' | sort; done "
        collector_cmd += ' > ' + cluster_result_file
        exit_code = os.system(collector_cmd)
        if exit_code != 0:
            logging.error('Failure in collecting cluster results')
            raise ValueError('Failure in collecting cluster results')

    if is_graph_clust:
        reference_label_file_knn = os.path.join(input_dir, 'FASTA/orig.fasta')
    else:
        reference_label_file_knn = os.path.join(input_dir, 'representatives.fa')
    # evaluating GraphClust requires separate reference label files for KNN and clustering, hence duplicate the label
    # files here:
    reference_label_file_clustering = reference_label_file_knn

    if graph_clust_all_sequences:
        # contains identifiers of representative sequences; one per line
        representatives_file = os.path.join(output_dir, 'representative_ids.txt')
        rep_identifiers = set()
        with open(representatives_file, 'r') as fhandle:
            for line in fhandle:
                rep_identifiers.add(line.rstrip())
        logging.info('GraphCluct with all sequences mode: detected {:d} representative sequences.'.format(
                len(rep_identifiers)))

        # modify cluster_result_file such that only representative sequences are retained. File looks as follows and
        # contains both representative sequences and other sequences:
        # 1 RF00017.5
        # 1 RF00017.6
        # 10 RF00027.0
        # 10 RF00027.1
        # 10 RF00027.10
        cluster_id_to_objects = collections.defaultdict(set)
        with open(cluster_result_file, 'r') as fhandle:
            for line in fhandle:
                cluster_id, object_id = line.rstrip().split(' ')
                if object_id in rep_identifiers:
                    cluster_id_to_objects[cluster_id].add(object_id)
        os.remove(cluster_result_file)

        with open(cluster_result_file, 'w') as fhandle:
            for i, cluster_name in enumerate(sorted(cluster_id_to_objects.keys())):
                for cluster_obj in cluster_id_to_objects[cluster_name]:
                    fhandle.write(str(i) + ' ' + cluster_obj + os.linesep)

        # write FASTA-like like containing identifiers of representative sequences. This file is later used to
        # infer the true clustering of the data set.
        reference_label_file_clustering = os.path.join(input_dir, 'pseudo.fa')
        with open(reference_label_file_clustering, 'w') as fhandle:
            for rep in rep_identifiers:
                fhandle.write('>' + rep + ' ' + rep + os.linesep)

    knn_out_path = os.path.join(output_dir, 'knn_analysis.tsv')
    cluster_out_path = os.path.join(output_dir, 'cluster_analysis.tsv')
    full_clustering_out_path = os.path.join(output_dir, 'cluster_result.tsv.gz')

    if not no_knn_analysis:
        logging.info('Processing feature vectors from file %s.' % feature_vector_file_knn)
        x = read_nspdk_feature(feature_vector_file_knn)

        logging.info('Loading reference labels from file %s.' % reference_label_file_knn)
        targets_list = read_fasta_targets(reference_label_file_knn)

        logging.info('Performing KNN analysis.')
        knn_result = knn_cross_validation(x, targets_list, kfold=cvfolds, dist_metric='cosine',
                                          non_stratified_cv=no_stratified_cv)

        logging.info('Writing KNN results to file %s.' % knn_out_path)
        with open(knn_out_path, 'w') as out_file:
            for result_tuple in knn_result:
                out_file.write('\t'.join(result_tuple) + os.linesep)

    logging.info('Evaluating clustering performance.')
    cluster_evaluation, cluster_result = eval_clustering_result(cluster_result_file, reference_label_file_clustering)

    logging.info('Writing clustering performance to file %s.' % cluster_out_path)
    with open(cluster_out_path, 'w') as out_file:
        out_file.write(cluster_evaluation)

    logging.info('Writing clustering results to file %s.' % full_clustering_out_path)
    with gzip.open(full_clustering_out_path, 'w') as out_file:
        for curr_cluster in cluster_result:
            out_file.write(';'.join(curr_cluster) + os.linesep)