#!/usr/bin/env python
import gzip
import itertools
import multiprocessing
import tarfile
import argparse
import os
import logging
import subprocess
import shutil

__author__ = 'Alexander Junge (alexander.junge@gmail.com)'


def is_valid_dir(arg):
    if not os.path.isdir(arg):
        parser.error("Given path %s is not a directory." % arg)
    return arg


def is_not_valid_path(arg):
    if os.path.exists(arg):
        parser.error("Given path %s does exist. Please specify a non-existing path.")
    return arg


def is_valid_path(arg):
    if not os.path.exists(arg):
        parser.error("Given path %s does not exist. Please specify an existing path.")
    return arg


def is_positive_int(arg):
    int_arg = None
    try:
        int_arg = int(arg)
    except ValueError:
        parser.error('Argument {} is not an integer.'.format(arg))
    if int_arg < 1:
        parser.error('Argument {} is not a positive integer.'.format(arg))
    return int_arg


def call_cmbuild(curr_cm_file_path, curr_align_file_path):
    cmbuild_call = 'cmbuild {} {}'.format(curr_cm_file_path, curr_align_file_path)
    logging.debug('Calling: ' + cmbuild_call)
    subprocess.check_call(cmbuild_call, shell=True)


def call_cmbuild_wrap(cm_align_file_pair):
    call_cmbuild(*cm_align_file_pair)


def call_cmconvert(curr_cm_file_path):
    cmconvert_call = 'cmconvert -1 {}'.format(curr_cm_file_path)
    logging.debug('Calling: ' + cmconvert_call)
    subprocess.check_output(cmconvert_call, shell=True)


def get_cmcompare_link_score(first_cm_path, second_cm_path):
    cmcompare_call = '{} {} {}'.format(args.cmcompare_path, first_cm_path, second_cm_path)
    logging.debug('Calling: ' + cmcompare_call)
    output = subprocess.check_output(cmcompare_call, shell=True)
    # CMCompare output looks like:
    # two.cm   three.cm     27.996     19.500 [..]
    # model 1, model 2, score 1, score 2 [..]
    output_split = output.split()
    first_score = float(output_split[2])
    second_score = float(output_split[3])
    curr_link_score = min(first_score, second_score)
    return curr_link_score


def get_cmcompare_link_score_wrap(cm_path_tuple):
    return get_cmcompare_link_score(*cm_path_tuple)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=
                                     """
                                     Compares a set of structural alignments using CMCompare [1] and writes their
                                     pairwise Link scores to a similarity matrix which can then be used for subsequent
                                     clustering.

                                     [1] Christian Hoener zu Siederdissen, & Ivo L. Hofacker (2010).
                                     Discriminatory power of RNA family models. Bioinformatics, 26(18), 453-459.
                                     doi:10.1093/bioinformatics/btq370
                                     """)
    parser.add_argument('input_dir', type=is_valid_dir,
                        help="""
                         Directory containing structural alignments to be compared.
                         Alignments must be in Stockholm format.
                         """)
    parser.add_argument('output_dir', type=is_not_valid_path,
                        help="""
                         Directory the output is written to.
                         """)
    parser.add_argument('-c', '--cmcompare_path', type=is_valid_path,
                        default='/home/users/choener/cmcompare/CMCompare',
                        help="""
                        Path of CMCompare binary. Default is location on RTH servers.
                        """)
    parser.add_argument('-t', '--thread_count', type=is_positive_int, default=1,
                        help="""
                            The number of threads to be started concurrently. Must be a positive integer.
                            """)
    parser.add_argument('-n', '--no_cmconvert', action='store_true',
                        help="""
                             Set this to skip CM conversion step from Infernal 1.1 to 1.0 format.
                             """)
    parser.add_argument('-d', '--debug', action='store_true', help='Enter debugging mode.')
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    try:
        os.makedirs(args.output_dir, mode=0774)
    except OSError:
        raise OSError('Failed to create output directory {}.'
                      'Make sure that output directory does not exist.'.format(args.output_dir))

    pool = multiprocessing.Pool(args.thread_count)

    input_file_paths = [os.path.join(args.input_dir, p)
                        for p in sorted(os.listdir(args.input_dir)) if os.path.isfile(os.path.join(args.input_dir, p))]
    logging.info('Found {:d} alignments in input directory'.format(len(input_file_paths)))

    if not all([os.path.basename(f).endswith('.stk') for f in input_file_paths]):
        raise ValueError('Detected files in input directory that do not end on .stk. Stopping.')

    # TODO make sure that consensus structure is available in each alignment and add it if missing

    logging.info('Generating covariance models for input alignments.')
    cm_output_dir = os.path.join(args.output_dir, 'cm')
    try:
        os.makedirs(cm_output_dir, mode=0774)
    except IOError:
        raise OSError('Failed to create CM output directory {}.'.format(cm_output_dir))

    align_count = float(len(input_file_paths))
    cm_file_paths = []
    for align_file_path in input_file_paths:
        cm_file_path = os.path.join(cm_output_dir, os.path.basename(align_file_path).rstrip('.stk') + '.cm')
        cm_file_paths.append(cm_file_path)
    pool.map(call_cmbuild_wrap, zip(cm_file_paths, input_file_paths))

    if args.no_cmconvert:
        pass
    else:
        pool.map(call_cmconvert, cm_file_paths)

    logging.info('Running CMCompare using {:d} processes.'.format(args.thread_count))
    cm_file_pairs = list(itertools.combinations_with_replacement(cm_file_paths, r=2))
    link_scores = pool.map(get_cmcompare_link_score_wrap, cm_file_pairs)

    # map each pair of CM file paths to the index of the corresponding link score in the link score list
    cm_file_pair_to_list_index = {}
    for list_index, cm_file_pair in enumerate(cm_file_pairs):
        cm_file_pair_to_list_index[cm_file_pair] = list_index

    # write link scores as similarity matrix to output folder specified by the user
    with gzip.open(os.path.join(args.output_dir, 'link.scores.tsv.gz'), 'w') as out_matrix_file:
        out_matrix_file.write('\t'.join([os.path.basename(cm_path) for cm_path in cm_file_paths]))
        out_matrix_file.write('\n')

        for cm_file1 in cm_file_paths:
            sep = ''
            for cm_file2 in cm_file_paths:
                k1 = (cm_file1, cm_file2)
                k2 = (cm_file2, cm_file1)
                if k1 in cm_file_pair_to_list_index:
                    link_score = str(link_scores[cm_file_pair_to_list_index[k1]])
                elif k2 in cm_file_pair_to_list_index:
                    link_score = str(link_scores[cm_file_pair_to_list_index[k2]])
                else:
                    raise ValueError('Could not find Link Score for covariance models {} and {}.'.format(cm_file1,
                                                                                                         cm_file2))
                out_matrix_file.write(sep + link_score)
                sep = '\t'
            out_matrix_file.write('\n')

    with tarfile.open(cm_output_dir + '.tar.gz', 'w:gz') as tarball:
        tarball.add(cm_output_dir, arcname=os.path.basename(cm_output_dir))
    shutil.rmtree(cm_output_dir)