#!/usr/bin/env bash
# Run this script inside Docker container when running the container with the following command:
# docker run -it -v "$CURRENT_WORKING_DIRECTORY"/cluster_evaluation:/cluster_evaluation -h dockersgeserver mmiladi/rnascclust:v"$IMAGE_VERSION"
#
# Inside the container, the present script can be executed by:
# bash rnascclust/bin/clustering/run_clustering_docker.sh >cluster_evaluation/run_clustering_docker.log 2>&1
#
perform_knn_analysis=true
docker_out="cluster_evaluation"
output_subdir_GC="RESULTS_GC"
output_subdir_GC_all_seq="RESULTS_GC_ALL_SEQ"
data_set_names=( "cliques-low" "cliques-medium" "cliques-high" "rfamome" )

for ((i=0; i < ${#data_set_names[@]}; i++))
do
    data_set=${data_set_names[$i]}
    out_dir_RSC="$docker_out/$data_set"
    out_dir_GC="$docker_out/$data_set/$output_subdir_GC"

    echo "Start RNAscClust for $data_set"
    mkdir -p "$out_dir_RSC"/RESULTS_SC
    RSC_out="$out_dir_RSC"/RESULTS_SC/log
    RSC_err="$out_dir_RSC"/RESULTS_SC/err

    date

    if [ "$perform_knn_analysis" = false ]; then
    	time /rnascclust/bin/run_rnascclust.py -f --no-knn-vector --alignments-dir "/rnascclust/data/$data_set/alignments/" --rep-fasta "/rnascclust/data/$data_set/representatives/representatives.fa" --predict-structure --result-path "/$out_dir_RSC/"/RESULTS_SC --local-config /rnascclust/Config/docker-sge.config --sge 1>"$RSC_out" 2>"$RSC_err"
    else
    	time /rnascclust/bin/run_rnascclust.py -f --alignments-dir "/rnascclust/data/$data_set/alignments/" --rep-fasta "/rnascclust/data/$data_set/representatives/representatives.fa" --predict-structure --result-path "/$out_dir_RSC/"/RESULTS_SC --local-config /rnascclust/Config/docker-sge.config --sge 1>"$RSC_out" 2>"$RSC_err"
    fi

    date

    echo "Start clustering evaluation"
    RSC_eval_dir="$out_dir_RSC"/RESULTS_SC/cluster_evaluation
    RSC_eval_log="$RSC_eval_dir"/log
    RSC_eval_err="$RSC_eval_dir"/err
    mkdir -p "$RSC_eval_dir"

    date

    if [ "$perform_knn_analysis" = false ]; then
    	echo "KNN analysis skipped."
        time python2 /rnascclust/bin/clustering/cluster_evaluation.py --no_knn_analysis "$out_dir_RSC"/RESULTS_SC "$RSC_eval_dir" 1>"$RSC_eval_log" 2>"$RSC_eval_log"
    else
        time python2 /rnascclust/bin/clustering/cluster_evaluation.py "$out_dir_RSC"/RESULTS_SC "$RSC_eval_dir" 1>"$RSC_eval_log" 2>"$RSC_eval_log"
    fi

    date
	# =========================================================================

    echo "Start GraphClust for $data_set"
    mkdir -p "$out_dir_GC"
    GC_out="$out_dir_GC"/log
    GC_err="$out_dir_GC"/err

    date

    # Note: Master script modified to not remove and keep the gspan file
    time /usr/local/user/GraphClust-0.7.6/bin/MASTER_GraphClust_debug.pl --fasta "/rnascclust/data/$data_set/representatives/representatives.fa" --root-dir "/$out_dir_GC/" --config /rnascclust/bin/GraphClust/config.cnf -sge -no-ssh 1>"$GC_out" 2>"$GC_err"

    date

    echo "Start clustering evaluation"
    GC_eval_dir="$out_dir_GC"/cluster_evaluation
    GC_eval_log="$GC_eval_dir"/log
    GC_eval_err="$GC_eval_dir"/err
	GC_gspan_file="$out_dir_GC"/GSPAN/1.group.gspan
    mkdir -p "$GC_eval_dir"

    date

    if [ "$perform_knn_analysis" = false ]; then
        echo "KNN analysis skipped."
        time python2 /rnascclust/bin/clustering/cluster_evaluation.py --graph_clust --no_knn_analysis "$out_dir_GC" "$GC_eval_dir" 1>"$GC_eval_log" 2>"$GC_eval_log"
    else
    	# 20 bit knn feature vector file should be generated first
    	bunzip2 $GC_gspan_file.bz2 1>"$GC_eval_log" 2>"$GC_eval_log"
    	cp $GC_gspan_file "$GC_gspan_file"_20bits

    	echo "Generate 20bit feature vectors with GraphClust."
    	time /usr/local/user/GraphClust-0.7.6/bin/NSPDK -R 3 -D 3 -b 20 -gt DIRECTED -of -fg "$GC_gspan_file"_20bits 1>"$GC_eval_log" 2>"$GC_eval_log"
    	date
        time python2 /rnascclust/bin/clustering/cluster_evaluation.py --graph_clust "$out_dir_GC" "$GC_eval_dir" 1>"$GC_eval_log" 2>"$GC_eval_log"
    fi

    date

done

echo "Producing Result figures and tables in paper."

date

if [ "$perform_knn_analysis" = false ]; then
    time python2 /rnascclust/bin/clustering/cluster_evaluation_manuscript.py --no_knn_analysis
else
    time python2 /rnascclust/bin/clustering/cluster_evaluation_manuscript.py
fi

date
