#!/usr/bin/env python

"""
    run_rnascclust
    ~~~~~~~~~~~~~
    Wrapper module to invoke the RNAAlignClust pipeline

@author:    Milad Miladi, Alexander Junge
            miladim@informatik.uni-freiburg.de, ajunge@rth.dk
            
    Sample invocation call command:
    $ python bin/run_rnascclust.py --alignments-dir data/cliques-low/alignments/ \
      --predict-structure  --result-path . --local-config Config/bilp13.config
"""

import os
import sys
import shutil
import argparse
import bz2
import glob
import subprocess
import string
from itertools import izip
from Bio import AlignIO, SeqIO
from Bio.Alphabet import IUPAC
from Bio.Align import MultipleSeqAlignment
import re
import time
import tempfile

import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


BIN_PATH = os.path.dirname(os.path.realpath(__file__))

import signal
__my_child_pid=None
import signal
def signal_handler(sig, frame):
        print('You pressed Ctrl+C!')
        print('child pid was', __my_child_pid)
        os.kill(__my_child_pid, signal.SIGTERM)
        sys.exit(0)


        
signal.signal(signal.SIGINT, signal_handler)

def is_valid_path(arg):
    if not os.path.exists(arg):
        parser.error("Given path %s does not exist" % arg)
    return arg

def is_parent_valid_path(arg):
    if not os.path.exists(os.path.dirname(arg.rstrip('/'))):
        parser.error("Parent path for %s does not exist" % arg)
    return arg

def is_valid_file(arg):
    if not os.path.isfile(arg):
        parser.error("Given file %s does not exist" % arg)
    return arg

# -------------
def remove_gap_columns( malign):
    ''' Remove all-gap columns from a biopython multiple-alignment
    Returns pruned multiple-alignment'''
    
    regx_gaps = '[-.~_]'  # valid gap symbols all be converted to "-"
    if (len(malign) == 0):
        return
    for c in reversed(range(len(malign[0]))):
        if len(re.sub(regx_gaps, '', malign[:, c])) == 0:
            malign = malign[:, 0:c] + malign[:, c + 1:]  # concat left & right side of column c
    
    return malign

# End def remove_gap_columns

def copy_representatives(represent_fasta, path_stks, target_dir):
    ''' Put a copy of each representative sequence beside its alignment file,
    the copy has extension .rep_fa'''
    print represent_fasta
    with open(represent_fasta, "r") as in_fasta_handle:
        rep_recs = list(SeqIO.parse(in_fasta_handle, "fasta"))
    
    rep_descs = [rec.description for rec in rep_recs]
#     with open()
    from Bio.SeqRecord import SeqRecord
    from Bio.Seq import Seq
    for rep in rep_recs:
        desc = rep.description
        seq_id, subaln_id = desc.split()
        if "_" in seq_id: # TODO check smarter for RFXXX_XYG11-11 ..
            seq_id = seq_id.split("_")[1]
#         print seq_id, subaln_id

        out_rep = "{}/{}.rep_fa".format(target_dir, subaln_id) 
        with open (out_rep, "w") as out_fa_handle:
#             SeqIO.write( SeqRecord(rep.seq, id=rep.id, description=""), out_fa_handle, "fasta")
            out_fa_handle.write(">{}\n{}".format(seq_id, str(rep.seq)))

def convert_fastas_to_stks(path_fastas, target_dir):
    if os.path.isdir(path_fastas) is False:
        logger.error("fasta path not found: {}".format(path_fastas))
    fasta_files = glob.glob(path_fastas+"/*.fa")
    os.mkdir(target_dir)

    # Collect first seqs of all alignments
    for fasta_file in fasta_files:
        in_fasta_handle = open (fasta_file, 'r')
        #dependent on the assumption that read() keeps the input file order
        # TODO: Check alphabet
        alignment = AlignIO.read(in_fasta_handle, "fasta", alphabet=IUPAC.ambiguous_rna)  
        assert (len(alignment) > 0)
        in_fasta_handle.close()

        out_stk_file = target_dir + "/" + os.path.basename(fasta_file).replace(".fa", ".stk")
        print out_stk_file
        with open(out_stk_file, 'w') as out_stk_handle:
            AlignIO.write(alignment, out_stk_handle, "stockholm")
    
    
def get_alignments(path_stks, min_class_size, target_dir):
    if os.path.isdir(path_stks) is False:
        logger.error("get_alignments: stk path not found: {}".format(path_stks))
    
    os.mkdir(target_dir)

    nonfiltered_stk_files = glob.glob(path_stks+"/*.stk")
    # sort files such that we get the order as bash ls command. RF00005.0, RF00005.10, RF00005.11, RF00005.1
# TODO: Sort files RF00005.0, RF00005.1, ..., RF00005.10, RF00005.11, 
    alphabet = string.printable  # 0-9A-Za-z_-. 
    nonfiltered_stk_files = sorted( nonfiltered_stk_files, 
                                    key=lambda word: [alphabet.index(c) for c in word+"..."])

    # Count size of the each class, stk_file base name must be class label
    labels_count = dict()
    for stk_file in nonfiltered_stk_files:
        label = os.path.basename(stk_file).split('.')[0] # Filename ohne extension!
        labels_count[label] = labels_count.get(label, 0) + 1  # Initialize/increment key value
    
    # Extract alignments that their reference label contains minimum number of alignments
    stk_files = [stk for stk in nonfiltered_stk_files 
                 if labels_count[os.path.basename(stk).split('.')[0]] >= min_class_size]
    
    print "Stk files reduced from", len(nonfiltered_stk_files), len(stk_files)
    
    # Copy filtered files after removing full-gap columns into target directory under RESULTS_AC
    for f in stk_files:
        target_f = '{}/{}'.format(target_dir, os.path.basename(f))
        with open(f) as stk_in, open(target_f, 'w') as stk_out:
            alignment = AlignIO.read(stk_in, "stockholm", alphabet=IUPAC.ambiguous_rna)  
            assert (len(alignment) > 0)
            alignment = remove_gap_columns(alignment)
            AlignIO.write(remove_gap_columns(alignment), stk_out, "stockholm")

#         shutil.copy(f, target_dir)
        
    copied_stk_files = glob.glob(target_dir+"/*.stk")
    return sorted(copied_stk_files,
                  key=lambda word: [alphabet.index(c) for c in word+"..."])

    
    
    
def generate_representatives(stk_files, out_fasta, keyword=None):
    '''Extracts a fasta file of representatives from all the alignments 
    available under the given path.
    Assumes first sequence of the alignment is the representative of the alignment
    keyword: If given the sequence id containing keyword is extracted as the representative. 
             Key must uniquely identify the representative sequence for all the multiple alignments
    '''
    
    assert (keyword is None) or len(keyword) > 0 
    # Collect first seqs of all alignments
    alignment_list = []        
    for stk_file in stk_files:
        stk_handle = open (stk_file, 'r')
        #dependent on the assumption that read() keeps the input file order
        # TODO: Check alphabet
        alignment = AlignIO.read(stk_handle, "stockholm", alphabet=IUPAC.ambiguous_rna)  
        assert (len(alignment) > 0)
        if keyword is None:
            
            rep = alignment[0]
        else:
            key_hits = [seq for seq in alignment if keyword in seq.id ]
            if len(key_hits) == 0:
                raise RuntimeError("keyword:{} not matched any sequence in file:{}".
                                   format(keyword, stk_file))
            elif len(key_hits) > 1:
                raise RuntimeError("keyword:{} matched {}>one sequences in file:{}".
                                   format(keyword, len(key_hits), stk_file))
            else:
                rep = key_hits[0]
             
        # TODO remove gaps
        rep.description = "{} {}".format(rep.id, 
                                         os.path.splitext(os.path.basename(stk_file))[0])
        gap_chars = '[-.~_]'  # valid gap symbols all be converted to "-"
        for gap in gap_chars:
            rep.seq = rep.seq.ungap(gap)

        alignment_list.append(rep)
        stk_handle.close()
    
    # Write single line sequences inside fasta file 
    from Bio.SeqIO.FastaIO import FastaWriter
    out_handle = open(out_fasta, 'w')
    writer = FastaWriter(out_handle, wrap=0)
    writer.write_file(alignment_list)
    out_handle.close()
#     with open(out_fasta, "w") as output_handle:
#         SeqIO.write(alignment_list, output_handle, "fasta")
#     
    logger.info("{} representatives were extracted".format(len(alignment_list)))
    
def remove_gspan_basepairs(gspan_f):
    ''' Remove secondary structure basepairs from the gspan graph'''
    gspan_f_nobp = gspan_f.replace ('.gspan', '_path_nabst.gspan')
    with open(gspan_f) as gspan_in, open(gspan_f_nobp, 'w') as gspan_out:
        for line in gspan_in:
            if line.isspace():
                gspan_out.write(line)
                continue
            # TODO: remove stack vertices as well
            splits = line.split()
            if splits[0] == 'e':
                assert(len(splits) > 3)
                if (int)(splits[2])-(int)(splits[1]) == 1: #backbone graph should be added
                    gspan_out.write(line)
                elif  abs((int)(splits[2])-(int)(splits[1])) > 1:
                    if ("#" in splits[3]) or ("^" in splits[3]) or ("@" in splits[3]): # abstract edges are kept
                        gspan_out.write(line)
                    else:
                        print "skipped: ", line
                elif splits[3] == 'p': #stack node
                    print "skipped: ", line

                else:
                    raise RuntimeError('Parsing gspan file edge expected with node difference more than one found: {}'.format(line))
            else: # non edge lines
                gspan_out.write(line)
                
    shutil.copyfile(gspan_f, gspan_f+'.withbps') # Copying can be avoided

    shutil.copyfile(gspan_f_nobp, gspan_f) # Copying can be avoided
    logger.info('basepair edges were removed from the gspan file')

def do_evaluation(results_dir):
    print "Start clustering evaluation within pipeline"
    RSC_eval_dir=results_dir + 'cluster_evaluation'
    if os.path.isdir(RSC_eval_dir):
        shutil.rmtree(RSC_eval_dir)
    os.mkdir(RSC_eval_dir)    
    
    RSC_eval_log=RSC_eval_dir + '/log'
    RSC_eval_err=RSC_eval_dir + '/err'
    
    if args.no_knn_vector is True:  # Generate smaller vector for knn classification analysis
        os.system('time python2 ' + BIN_PATH + 
                  '/clustering/cluster_evaluation.py --no_knn_analysis ' + 
                  results_dir + " " + RSC_eval_dir + " 1> " + RSC_eval_log + " 2> " + RSC_eval_log)
    else:
        os.system('time python2 ' + BIN_PATH + 
                  '/clustering/cluster_evaluation.py ' + 
                  results_dir + " " + RSC_eval_dir + " 1> " + RSC_eval_log + " 2> " + RSC_eval_log)
    #

def filter_with_blastclust(stk_files, represent_fasta):
    ''' Runs blastclust on representatives and filters highly sequence-similar ones
        returns: list of filtered_stk_files , filtered_represent_fasta
    '''
    
#     if not os.path.isdir(blast_outdir):
#         raise RuntimeError ("Given path %s does not exist" % blast_outdir)
#     
#         
    from subprocess import Popen, PIPE
    blast_out_file = represent_fasta.replace('.fa','').replace('.fasta','') + '.blast.out' 
    cmd = " ".join(['blastclust',
                    '-p F -b T -W 20 -S 90 -L 0.9', 
                    '-i' , represent_fasta, 
                    '-o', blast_out_file ,
                    ])
    
    p = Popen(cmd, stdin=PIPE, shell=True, stdout=PIPE, stderr=PIPE)
    out, err = p.communicate()
    if err:
        print "Error in calling locarna_rnafold_pp\n", out, err
    print out
    
    # Read blastclust outfile and extract clusters with more than one entry
    blast_hits = dict()
    blast_refs_to_filter = set()

    with open(blast_out_file) as blast_out_handle:
        for line in blast_out_handle:
            splits = line.split()
            if len(splits) < 2:
                continue
            
            for split in splits[1:]:
                blast_hits[split] = splits[0]
    
    print blast_hits
    from Bio.SeqIO.FastaIO import FastaWriter
    filtered_represent_fasta = represent_fasta.replace('.fa','').replace('.fasta','') + '.blast_filtered.fa' 
    with open(represent_fasta, "r") as in_fasta_handle, open(filtered_represent_fasta, "w") as out_fasta_handle:
        for rec in SeqIO.parse(in_fasta_handle, "fasta"):
            if rec.id in blast_hits:
                blast_refs_to_filter.add(rec.description.split()[1])
                continue
            SeqIO.write(rec, out_fasta_handle, 'fasta')
            
    print blast_refs_to_filter
    filtered_stk_files = list()
    for stk_file in stk_files:
        ref_stk = os.path.splitext(os.path.basename(stk_file))[0]
        if ref_stk in blast_refs_to_filter:
            continue
        else:
            filtered_stk_files.append(stk_file)
    
    logger.info('Blastclust reduced and filtered input entries from {} to {}'.format(len(stk_files), 
                                                                         len(filtered_stk_files)))
    return filtered_stk_files, filtered_represent_fasta
    


def sanity_check(stk_files, represent_fasta):
    '''Verifies alignments and the representative fasta have same size and order'''
    
    with open(represent_fasta, "r") as in_fasta_handle:
        fa_descs = [rec.description for rec in SeqIO.parse(in_fasta_handle, "fasta")]
    
    # Count size of the each class, stk_file base name must be class label
    labels_count = dict()
    for stk_file in stk_files:
        label = os.path.basename(stk_file)
        labels_count[label] = labels_count.get(label, 0) + 1  # Initialize/increment key value
        
        # Check alignments can be parsed by biopython and are not empty
        stk_handle = open (stk_file, 'r')
        #dependent on the assumption that read() keeps the input file order
        alignment = AlignIO.read(stk_handle, "stockholm")
        assert (len(alignment) > 0)
        stk_handle.close()
    #end for 
    
    if len(fa_descs) != len(stk_files):
        logger.error("Number of alignment files:{} and sequences of representative "\
        "fasta:{} must match ".format(len(stk_files), len(fa_descs)))
        raise AssertionError
    
    
    # Fasta descriptors (second entrY OF each fasta header) should match to stk filename 
    for desc, file in izip(fa_descs, stk_files):
        file = os.path.splitext(os.path.basename(file))[0]
        
        if (len(desc.split()) < 2) or (desc.split()[1] != file):
            logger.error("Mismatch between representative fasta and alignment files")
            logger.error("Fasta-descriptor:{} alignment-file:{}".format(desc, file))
            raise AssertionError
    #end for
    logger.info("Sanity check of the representatives passed!")
#         stk_handle = open (stk_file, 'r')
        #dependent on the assumption that read() keeps the input file order
#         alignment = AlignIO.read(stk_handle, "stockholm") 

# =========================================================================
#
# =========================================================================
print "Args: ", sys.argv
parser = argparse.ArgumentParser()
'''
        Options from perl script not integrated
        
                         "stdout"      => \$i_stdout,
                         "cutoff-ratio-consensus=i" => \$i_cutoff_cons,
                         "cutoff-ratio-MFE=i" => \$i_cutoff_mfe,
                         "read-stockholm" => \$i_stockholm,
                         "read-clustal" => \$i_clustal,
                         "read-aln-fasta" => \$i_aln_fasta,
                         "threshold-reliability=f" => \$i_threshold_reli,
                         );

'''
argparse.ArgumentParser(description="""RNAAlignClust Pipeline """)
                
parser.add_argument('--alignments-dir', "-i", type=is_valid_path, required=True,
                    help='Path to the directory containing stockholm files')

parser.add_argument('--result-path', type=is_parent_valid_path, required=True,
                    help='Path to creat output directory RESULTS_AC/ under\
                    it stores intermediate and final results ')
parser.add_argument('--no-knn-vector', action='store_true',
                    help='set this to avoid create an extra feature vector \
                    of size 20bits necessary for KNN analysis')
parser.add_argument('--fasta-alignments', action='store_true', default=False,
                    help='Set if alignments in alignments-dir are in fasta alignment format ')

parser.add_argument('--local-config', type=is_valid_file,
                    help='if a local configuration is to be used, path to the \
                    local configuration file')

parser.add_argument('--GC-config', type=is_valid_file,
                    help='if a custom GC(i.e. iterative clustering) configuration is to be used, path to the \
                    GC configuration file')

parser.add_argument('--sge-parallel', action='store_true', default=False,
                    help='Run in parallel multicore/SGE mode ')
parser.add_argument('--host-sge', default='docker',
                    help='SGE host name, availble options are: "docker", "uibifr" ')

parser.add_argument('--rep-fasta', type=is_valid_file, required=False,
                    help='A fasta file containing one representative sequence \
                    from each alignment under the input-directory.\
                    Representative could be e.g. the human sequence of each alignment')

parser.add_argument('--representative-keyword', type=str, required=False,
                    help='A keyword (as a substring of sequence ids header) that can uniquely identify \
                    representative id of each multiple alignment')

parser.add_argument('--temp-path', type=is_valid_path, required=False,
                    default='/tmp/',  # TODO: find the host temp path in generic way
                    help='Path to store intermediate temporary direcotry named TEMP_AC ')

parser.add_argument('--structure-is-given', action='store_true', default=False,
                    help='Do not use PETfold to predict the secondary consensus "\
                    structure of each alignment')

parser.add_argument('--mfe', action='store_true', default=False,
                    help='Use MFE of each sequence as encoded graph,\
                     this option ignores alignment and evolution information\
                      and is mainly for evaluation purposes')
parser.add_argument('--path-graph', action='store_true', default=False,
                    help='Use only path graph of each sequence as encoded graph,\
                     this option ignores alignment and evolution information\
                      and is mainly for evaluation purposes')
parser.add_argument('--path-abstract-graph', action='store_true', default=False,
                    help='Use both path graph and abstract of each sequence as encoded graph,\
                     this iscurrently for evaluation purposes')

parser.add_argument('--only-representative', action='store_true', default=False,
                    help='Only encode the given representative as graph and ignore other sequences')


parser.add_argument('--max-selection-size', type=int, default=1,
                    help='Max number of sequences, including representative, to be encoded as graph\
                    size one is equivalent to only-representative option')

parser.add_argument('--perform-evaluation', action='store_true', default=False,
                    help='Perform knn/clustering evaluation after running the pipeline')

parser.add_argument('--only-evaluation', action='store_true', default=False,
                    help='Perform only knn/clustering evaluation on the available results')

parser.add_argument('--threshold-reliability', type=float, 
                    help='Petfold reliability threshold')

parser.add_argument('--predict-structure', dest='structure_is_given', action='store_false',
                    help='Use PETfold to predict the secondary consensus \
                    structure of each alignment')

parser.add_argument('--keep-temp', '-k', action='store_true',
                    help='Keep files after run finished')

parser.add_argument('--min-class-size', type=int, default = 3,
                    help='Discard families where number of subalignments is less than min')

parser.add_argument('--force', '-f', action='store_true',
                    help='Force overwriting to non-empty results-dir.\
                    WARNING: Entire results-dir pass will be removed recursively!')

parser.add_argument('--skip-graph-encoding', action='store_true',
                    help='Skip gspan graph generation and use current files')

parser.add_argument('--prefilter-blast', action='store_true',
                    help='Prefilter input sequences using blastclust')

parser.add_argument('--debug', '-d', action='store_true',
                    help='Debug mode, verbose information logged')

parser.add_argument('-job-id', help="dummy param")

args = parser.parse_args()



# =========================================================================
# 
# =========================================================================
if args.max_selection_size == 1:
    args.only_representative = True
elif args.max_selection_size > 1:
    if args.only_representative is True:
        raise RuntimeError("Error: max-selection-size cannot be more than 1, if only-representative is set")
else:
    raise RuntimeError("Error: max-selection-size cannot be less than 1")    
        
if args.mfe and args.path_graph:
    raise RuntimeError("Error at most one of MFE and PATH options can be set")
if args.path_graph and args.path_abstract_graph:
     raise RuntimeError("Error at most one of path-graph and path-abstract-graph can be set")

if args.rep_fasta and args.representative_keyword:
    raise RuntimeError("Error at most one of rep-fasta and representative-keyword options can be set")

# if args.mfe is True:
#     results_SC_dir = args.result_path.rstrip('/') + '/RESULTS_SC_MFE'
# elif  args.path_graph is True:
#     results_SC_dir = args.result_path.rstrip('/') + '/RESULTS_SC_PATH'
# elif  args.path_abstract_graph is True:
#     results_SC_dir = args.result_path.rstrip('/') + '/RESULTS_SC_PATH_ABSTRACT'
# else:
#     results_SC_dir = args.result_path.rstrip('/') + '/RESULTS_SC'
#     
# results_SC_dir += '-max-{}/'.format(args.max_selection_size)

results_SC_dir = args.result_path.rstrip('/') + "/"

    
if args.only_evaluation:
    do_evaluation(results_SC_dir)
    sys.exit()

try:
    if os.path.isdir(results_SC_dir):
        if args.skip_graph_encoding:
            logger.info("Use previously computed gspan grpahs ")
    	elif os.listdir(results_SC_dir) and len(os.listdir(results_SC_dir)) > 0:
            if args.force is True:
                logger.info("Cleaning up the results directory")
                shutil.rmtree(results_SC_dir)
                logger.info("Creating the results directory")
                os.mkdir(results_SC_dir)    
            else:
                logger.error("The specified results dir is not empty,"+ 
                             "please specify an empty directory or add "+ 
                             "--f to cleanup the directory: %s" % results_SC_dir)
                raise IOError()
    else:
        logger.info("Creating the results directory")
        os.mkdir(results_SC_dir)    
except Exception as ex:
    logger.error("Unable to modify/remove contents of the result directory with path %s" % results_SC_dir)
    raise ex


temp_AC_dir = tempfile.mkdtemp(prefix='TEMP_AC', dir=args.temp_path)
 
try:
    if os.path.isdir(temp_AC_dir):
        shutil.rmtree(temp_AC_dir)
    os.mkdir(temp_AC_dir)    
except:
    logger.error("Unable to modify/remove content of the temp directory with path %s" % temp_AC_dir)
    raise

# Extract list of alignment files, filter them if needed

if args.fasta_alignments is True:
    stk_tmp_dir = results_SC_dir + "INTERMEDIATE_INPUT/"
    convert_fastas_to_stks(args.alignments_dir, stk_tmp_dir)
    extracted_alignments = get_alignments(stk_tmp_dir, args.min_class_size, 
                                      results_SC_dir + "INPUT/" )
    print "extracteds are:"
    print extracted_alignments
else:
    extracted_alignments = get_alignments(args.alignments_dir, args.min_class_size, 
                                      results_SC_dir + "INPUT/" )
represent_fasta_file = results_SC_dir + 'representatives.fa' # Todo extract human

if os.path.isfile( represent_fasta_file):
    os.remove(represent_fasta_file)
# If representatives are provided by user
if args.rep_fasta:

    try:
        shutil.copy(args.rep_fasta, represent_fasta_file)
    except:
        logger.error("Unable to copy provided fasta into results directory with path %s" % represent_fasta_file)
        raise
        
elif args.representative_keyword:
    generate_representatives(extracted_alignments, represent_fasta_file, keyword=args.representative_keyword)
    
else:  # otherwise extract a sequence of each alignment as rep
    logger.warning("Fasta file for representative sequences is not provided,\
        considering the first sequence of each alignment as its representative!")
    generate_representatives(extracted_alignments, represent_fasta_file)

# Verify the provided or generated representatives against the alignment files 
sanity_check(extracted_alignments, represent_fasta_file )

if args.prefilter_blast is True:
    extracted_alignments, represent_fasta_file = filter_with_blastclust(extracted_alignments, represent_fasta_file)
    sanity_check(extracted_alignments, represent_fasta_file )


copy_representatives(represent_fasta_file, args.alignments_dir, results_SC_dir + "INPUT/")

# =========================================================================
#
# =========================================================================
logger.info("Generate Graph Encoded Alignments")
''' Example:
perl bin/alignment_to_secondary_structures.pl -input-aln-file=RF02106_filtered.2 
-debug -result-dir=. -tmp-dir=./tmp/ -cutoff-ratio-consensus=20 -cutoff-ratio-MFE=20 
-use-petfold -read-stockholm -threshold-reliablity=0.001 --local-config=Config/bilp13.config
'''

perl_args = ["--input-dir", results_SC_dir+"INPUT/",
              "--result-dir", results_SC_dir,
              "--tmp-dir",  temp_AC_dir,
              "--use-petfold" if not args.structure_is_given else "",
              "--mfe" if args.mfe else "",
              "--path-graph" if args.path_graph else "",
              "--path-abstract-graph" if args.path_abstract_graph else "",
              "--max-selection-size=%d" % args.max_selection_size,
              "--mfe" if args.mfe else "",
              "--threshold-reliability=%f" % (args.threshold_reliability) if args.threshold_reliability else "",
              "--keep-temp" if args.keep_temp else "",
              "--debug" if args.debug else "",
              "--sge" if args.sge_parallel else "",
              "--read-stockholm"
              ]
if args.local_config:
    perl_args += ["--local-config", args.local_config]

print perl_args


# Run gspan graph encoder scripts
if args.skip_graph_encoding:
    logger.info("Use previously computed gspan grpahs ")
else:
    with open(results_SC_dir+'phase1-graph_encoder.stdout.log', 'w') as logfile:
        encoder_pipe = subprocess.Popen(["perl", BIN_PATH + "/iterate_over_alignments.pl"] + perl_args, 
                     stdout=logfile#subprocess.PIPE#, stderr=subprocess.PIPE #logfile,                    preexec_fn=os.setsid
                     )
        __my_child_pid = encoder_pipe.pid
        encoder_pipe.communicate()
        
    

wait_count = 0
while (len(glob.glob(results_SC_dir + "GSPAN/*.bz2")) < len(extracted_alignments)):
    wait_count += 1
    logger.warning("Waiting for all encoded graphs to be synchronized with the host file system {}".format(wait_count))
    time.sleep(10)

    if wait_count > 12:
        break
    
    
if len(glob.glob(results_SC_dir + "GSPAN/*.bz2")) == 0:
    logger.error("No gspan Encoded Graph was generated from the first phase, Exit..")
    sys.exit()
elif len(glob.glob(results_SC_dir + "GSPAN/*.bz2")) < len(extracted_alignments):
    logger.error("Not all gspan Encoded Graph was generated from the first phase," +
                 " expected:{} acquired:{} Exit..".format(len(extracted_alignments), 
                                                         len(glob.glob(results_SC_dir + "GSPAN/*.bz2")) ))
    sys.exit()

logger.info("Graph encoding Alignment finished")

gspan_file = results_SC_dir + "encoded_alignments.gspan"
gspan_file_20b = results_SC_dir + "encoded_alignments_20bits.gspan"

if os.path.isfile(gspan_file):  # Remove old gspan file if exist
    os.remove(gspan_file)
    

# Decompress the bz2 gspan files and write them into a single file
with open(gspan_file, 'wb') as outfile:
    for bz2file in sorted(glob.glob(results_SC_dir + "GSPAN/*.bz2")):
        with bz2.BZ2File(bz2file, 'rb') as infile:
            for chunk in iter(lambda : infile.read(100 * 1024), b''):
                outfile.write(chunk)

# Remove secondary structure basepair edges from the graph
if args.path_abstract_graph:
    remove_gspan_basepairs(gspan_file)

# ==================================== =====================================
# NSPDK creates sparse feature vectors
# =========================================================================
logger.info("Generate Sparse Feature Vector of the Graph Encoded Alignments")
nspdk_path = BIN_PATH + '/GraphClust/NSPDK'
nspdk_args =  " -R 3 -D 3  -gt DIRECTED -of -fg %s" % gspan_file # Todo use config file params
print [nspdk_path, nspdk_args ]
vectorizer_pipe = subprocess.call (nspdk_path + nspdk_args, shell=True)

if args.no_knn_vector is False:  # Generate smaller vector for knn classification analysis
    
    logger.info("Generate 20bits Sparse Feature Vector of the Graph Encoded Alignments")
    nspdk_path = BIN_PATH + '/GraphClust/NSPDK'
    shutil.copyfile(gspan_file, gspan_file_20b)
    nspdk_args =  " -R 3 -D 3 -b 20 -gt DIRECTED -of -fg %s" % gspan_file_20b # Todo use config file params
    print [nspdk_path, nspdk_args ]
    vectorizer_pipe = subprocess.call (nspdk_path + nspdk_args, shell=True)
    


# =========================================================================
# Hybrid pipeline Candidate CLuster Selection and Iterative clustering
# =========================================================================
logger.info("Candidate CLuster Selection and Iterative clustering")

hybrid_path = BIN_PATH + '/GraphClust/MASTER_Hybrid_debug.pl'
if args.GC_config:
    GC_config_file = args.GC_config 
elif args.host_sge == 'docker':
    GC_config_file = BIN_PATH + '/GraphClust/config.cnf'
elif args.host_sge == 'uibifr':
    GC_config_file = BIN_PATH + '/GraphClust/config_uibifr.cnf'
else:
    if args.sge_parallel:
        print "WARNING: sge host name unknown, sge submission is prone to fail"
        GC_config_file = BIN_PATH + '/GraphClust/config.cnf'
hybrid_args = ['--fasta',  represent_fasta_file,
               '--root-dir', results_SC_dir,
               '--feature-vector', gspan_file + ".feature_bin",
               '--config', GC_config_file,
               '--no-ssh',
               '--sge'  if args.sge_parallel else ""
               ]

print ["perl", hybrid_path] + hybrid_args

hybrid_pipe = subprocess.call (["perl", hybrid_path] + hybrid_args )

cluster_dirs = results_SC_dir+"/RESULTS/[0-9]*"
# One liner collector of the clustering results
collector_cmd = "for c in " + cluster_dirs + "; do cat $c/cluster.all | awk -v c=$(basename $c) '{print c,$11}' | sort; done "
collector_cmd += " > " +results_SC_dir+"clustering_results_overall.txt"
exit_code = os.system(collector_cmd)
if (exit_code != 0):
    logger.error("Failure in collecting cluster results")
    raise


# =================
# Run evaluation within pipeline
# =================
if args.perform_evaluation is True:
    do_evaluation(results_SC_dir)
    



